import os
import platform

import re
import logging
import json
from structs.data import Cell, Table, HFTabularDataset


logger = logging.getLogger(__name__)

class HybridQA(HFTabularDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _load_split(self, split):
        question_file_map = {
            "train": "train.json",
            "dev": "dev.json",
            "test": "test.json"
        }

        file_base_path = 'datasets/HybridQA'
        logger.info(f"Loading HybridQA - {split}")
        question_file_path = os.path.join(file_base_path, question_file_map[split])

        with open(question_file_path, 'r', encoding='utf-8') as f:
            question_data_list = json.load(f)

        data = []

        for question_data in question_data_list:
            properties_info = {}
            question = question_data['question']
            table_id = question_data['table_id']

            if 'question_postag' in question_data:
                properties_info['question--question_postag'] = str(question_data['question_postag'])


            if platform.system() == 'Windows':
                illegal_chars = r'[\\/:\*\?"<>|]'
                table_id = re.sub(illegal_chars, '_', table_id)

            with open(f'{file_base_path}/tables_tok/{table_id}.json', 'r', encoding='utf-8') as f:
                table_data = json.load(f)

            with open('{}/request_tok/{}.json'.format(file_base_path, table_id), 'r', encoding='utf-8') as f:
                requested_document = json.load(f)

            table_data_headers = [aa[0] for aa in table_data['header']]


            table_data_contents = []
            for row in table_data['data']:
                table_data_row_contents = []
                for cell in row:
                    table_data_cell_content = cell[0]
                    for i, link_data in enumerate(cell[1]):
                        if i == 0:
                            table_data_cell_content += '##[HERE STARTS THE HYPERLINKED PASSAGE]##'
                        table_data_cell_content += ("[HYPERLINKED PASSAGE "+str(i+1)+"]: "+requested_document[link_data])
                    table_data_row_contents.append(table_data_cell_content)
                table_data_contents.append(table_data_row_contents)

            if 'url' in table_data:
                properties_info['table--url'] = str(table_data['url'])
            if 'title' in table_data:
                properties_info['table--title'] = str(table_data['title'])
            if 'section_title' in table_data:
                properties_info['table--section_title'] = str(table_data['section_title'])
            # if 'section_text' in table_data:
            #     properties_info['table[section_text]'] = str(table_data['section_text'])
            # if 'intro' in table_data:
            #     properties_info['table[intro]'] = str(table_data['intro'])


            data.append({
                "question": question,
                "table": {
                    "header": table_data_headers,
                    "content": table_data_contents
                },
                "properties": properties_info
            })
        self.data[split] = data
        self.dataset_info = {
            "citation": "111",
            "description": "111",
            "version": "111",
            "license": "111"
        }

    def load_split_test(self, split):
        self._load_split(split)

    def prepare_table(self, entry):
        t = Table()
        t.is_linked = True
        t.type = "default"
        t.default_question = entry["question"]

        t.props = entry['properties']

        for header_cell in entry["table"]["header"]:
            c = Cell()
            c.value = header_cell
            c.is_col_header = True
            t.add_cell(c)
        t.save_row()

        for row in entry["table"]["content"]:
            for cell in row:
                c = Cell()
                c.value = cell
                t.add_cell(c)
            t.save_row()
        return t
if __name__ == '__main__':
    HybridQA().load_split_test("train")


import argparse, json
from Read import GenerateTrainer, MRCTrainer

def read_config(path):
    with open(path, 'r') as file:
        config = json.load(file)
    return config

if __name__ == '__main__':
    parser  = argparse.ArgumentParser()
    parser.add_argument('--train_data_path', type=str, default='/home/lfy/S3HQA1/Data/HybridQA/train.row.json')
    parser.add_argument('--dev_data_path', type=str, default='/home/lfy/S3HQA1/Data/HybridQA/dev.row.json')
    parser.add_argument('--dev_reference', type=str, default='/home/lfy/S3HQA1/Data/HybridQA/dev_reference.json')
    parser.add_argument('--WikiTables', type=str, default='/home/lfy/S3HQA1/Data/HybridQA/WikiTables-WithLinks')
    parser.add_argument('--is_train', type=int, default=1)
    parser.add_argument('--output_path', type=str, default='./try')
    parser.add_argument('--mode', choices=['mrc', 'generate'])
    parser.add_argument('--plm', type=str, default='/home/lfy/S3HQA1/PTM/bart-large')

    args = parser.parse_args()
    training_config = read_config(args.config_file)

    if args.mode == 'generate':
        mytrainer = GenerateTrainer(args)
        mytrainer.train(**training_config)
    elif args.mode == 'mrc':
        mytrainer = MRCTrainer(args)
        mytrainer.train(**training_config)
    
    


import os


working_dir = os.path.dirname(os.path.abspath(__file__))
workers = 2
bind = "210.75.240.136:18888"
accesslog = "gunicorn_access.log"
errorlog = "gunicorn_error.log"
loglevel = "info"
timeout = 300



"""
Utilities for working with the local dataset cache. This file is adapted from the AllenNLP library at
https://github.com/allenai/allennlp Copyright by the AllenNLP authors.
"""

import fnmatch
import json
import os
import re
import shutil
import sys
import tarfile
import tempfile
from collections import OrderedDict
from contextlib import contextmanager
from dataclasses import fields
from functools import partial, wraps
from hashlib import sha256
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union
from urllib.parse import urlparse
from zipfile import ZipFile, is_zipfile

import numpy as np
from tqdm.auto import tqdm

import requests
from filelock import FileLock

# from . import __version__
from .utils import logging


logger = logging.get_logger(__name__)  # pylint: disable=invalid-name

ENV_VARS_TRUE_VALUES = {"1", "ON", "YES"}
ENV_VARS_TRUE_AND_AUTO_VALUES = ENV_VARS_TRUE_VALUES.union({"AUTO"})

try:
    USE_TF = os.environ.get("USE_TF", "AUTO").upper()
    USE_TORCH = os.environ.get("USE_TORCH", "AUTO").upper()
    if USE_TORCH in ENV_VARS_TRUE_AND_AUTO_VALUES and USE_TF not in ENV_VARS_TRUE_VALUES:
        import torch

        _torch_available = True  # pylint: disable=invalid-name
        logger.info("PyTorch version {} available.".format(torch.__version__))
    else:
        logger.info("Disabling PyTorch because USE_TF is set")
        _torch_available = False
except ImportError:
    _torch_available = False  # pylint: disable=invalid-name

try:
    USE_TF = os.environ.get("USE_TF", "AUTO").upper()
    USE_TORCH = os.environ.get("USE_TORCH", "AUTO").upper()

    if USE_TF in ENV_VARS_TRUE_AND_AUTO_VALUES and USE_TORCH not in ENV_VARS_TRUE_VALUES:
        import tensorflow as tf

        assert hasattr(tf, "__version__") and int(tf.__version__[0]) >= 2
        _tf_available = True  # pylint: disable=invalid-name
        logger.info("TensorFlow version {} available.".format(tf.__version__))
    else:
        logger.info("Disabling Tensorflow because USE_TORCH is set")
        _tf_available = False
except (ImportError, AssertionError):
    _tf_available = False  # pylint: disable=invalid-name


try:
    USE_JAX = os.environ.get("USE_FLAX", "AUTO").upper()

    if USE_JAX in ENV_VARS_TRUE_AND_AUTO_VALUES:
        import flax
        import jax

        logger.info("JAX version {}, Flax: available".format(jax.__version__))
        logger.info("Flax available: {}".format(flax))
        _flax_available = True
    else:
        _flax_available = False
except ImportError:
    _flax_available = False  # pylint: disable=invalid-name


try:
    import datasets  # noqa: F401

    # Check we're not importing a "datasets" directory somewhere
    _datasets_available = hasattr(datasets, "__version__") and hasattr(datasets, "load_dataset")
    if _datasets_available:
        logger.debug(f"Successfully imported datasets version {datasets.__version__}")
    else:
        logger.debug("Imported a datasets object but this doesn't seem to be the 🤗 datasets library.")

except ImportError:
    _datasets_available = False

try:
    from torch.hub import _get_torch_home

    torch_cache_home = _get_torch_home()
except ImportError:
    torch_cache_home = os.path.expanduser(
        os.getenv("TORCH_HOME", os.path.join(os.getenv("XDG_CACHE_HOME", "~/.cache"), "torch"))
    )


try:
    import torch_xla.core.xla_model as xm  # noqa: F401

    if _torch_available:
        _torch_tpu_available = True  # pylint: disable=
    else:
        _torch_tpu_available = False
except ImportError:
    _torch_tpu_available = False


try:
    import psutil  # noqa: F401

    _psutil_available = True

except ImportError:
    _psutil_available = False


try:
    import py3nvml  # noqa: F401

    _py3nvml_available = True

except ImportError:
    _py3nvml_available = False


try:
    from apex import amp  # noqa: F401

    _has_apex = True
except ImportError:
    _has_apex = False


try:
    import faiss  # noqa: F401

    _faiss_available = True
    logger.debug(f"Successfully imported faiss version {faiss.__version__}")
except ImportError:
    _faiss_available = False

try:
    import sklearn.metrics  # noqa: F401

    import scipy.stats  # noqa: F401

    _has_sklearn = True
except (AttributeError, ImportError):
    _has_sklearn = False

try:
    # Test copied from tqdm.autonotebook: https://github.com/tqdm/tqdm/blob/master/tqdm/autonotebook.py
    get_ipython = sys.modules["IPython"].get_ipython
    if "IPKernelApp" not in get_ipython().config:
        raise ImportError("console")
    if "VSCODE_PID" in os.environ:
        raise ImportError("vscode")

    import IPython  # noqa: F401

    _in_notebook = True
except (AttributeError, ImportError, KeyError):
    _in_notebook = False


try:
    import sentencepiece  # noqa: F401

    _sentencepiece_available = True

except ImportError:
    _sentencepiece_available = False


try:
    import tokenizers  # noqa: F401

    _tokenizers_available = True

except ImportError:
    _tokenizers_available = False


try:
    import torch_scatter

    # Check we're not importing a "torch_scatter" directory somewhere
    _scatter_available = hasattr(torch_scatter, "__version__") and hasattr(torch_scatter, "scatter")
    if _scatter_available:
        logger.debug(f"Succesfully imported torch-scatter version {torch_scatter.__version__}")
    else:
        logger.debug("Imported a torch_scatter object but this doesn't seem to be the torch-scatter library.")

except ImportError:
    _scatter_available = False


default_cache_path = os.path.join(torch_cache_home, "transformers")


PYTORCH_PRETRAINED_BERT_CACHE = os.getenv("PYTORCH_PRETRAINED_BERT_CACHE", default_cache_path)
PYTORCH_TRANSFORMERS_CACHE = os.getenv("PYTORCH_TRANSFORMERS_CACHE", PYTORCH_PRETRAINED_BERT_CACHE)
TRANSFORMERS_CACHE = os.getenv("TRANSFORMERS_CACHE", PYTORCH_TRANSFORMERS_CACHE)

WEIGHTS_NAME = "pytorch_model.bin"
TF2_WEIGHTS_NAME = "tf_model.h5"
TF_WEIGHTS_NAME = "model.ckpt"
CONFIG_NAME = "config.json"
MODEL_CARD_NAME = "modelcard.json"

SENTENCEPIECE_UNDERLINE = "▁"
SPIECE_UNDERLINE = SENTENCEPIECE_UNDERLINE  # Kept for backward compatibility

MULTIPLE_CHOICE_DUMMY_INPUTS = [
    [[0, 1, 0, 1], [1, 0, 0, 1]]
] * 2  # Needs to have 0s and 1s only since XLM uses it for langs too.
DUMMY_INPUTS = [[7, 6, 0, 0, 1], [1, 2, 3, 0, 0], [0, 0, 0, 4, 5]]
DUMMY_MASK = [[1, 1, 1, 1, 1], [1, 1, 1, 0, 0], [0, 0, 0, 1, 1]]

S3_BUCKET_PREFIX = "https://s3.amazonaws.com/models.huggingface.co/bert"
CLOUDFRONT_DISTRIB_PREFIX = "https://cdn.huggingface.co"
PRESET_MIRROR_DICT = {
    "tuna": "https://mirrors.tuna.tsinghua.edu.cn/hugging-face-models",
    "bfsu": "https://mirrors.bfsu.edu.cn/hugging-face-models",
}


def is_torch_available():
    return _torch_available


def is_tf_available():
    return _tf_available


def is_flax_available():
    return _flax_available


def is_torch_tpu_available():
    return _torch_tpu_available


def is_datasets_available():
    return _datasets_available


def is_psutil_available():
    return _psutil_available


def is_py3nvml_available():
    return _py3nvml_available


def is_apex_available():
    return _has_apex


def is_faiss_available():
    return _faiss_available


def is_sklearn_available():
    return _has_sklearn


def is_sentencepiece_available():
    return _sentencepiece_available


def is_tokenizers_available():
    return _tokenizers_available


def is_in_notebook():
    return _in_notebook


def torch_only_method(fn):
    def wrapper(*args, **kwargs):
        if not _torch_available:
            raise ImportError(
                "You need to install pytorch to use this method or class, "
                "or activate it with environment variables USE_TORCH=1 and USE_TF=0."
            )
        else:
            return fn(*args, **kwargs)

    return wrapper


# docstyle-ignore
def is_sklearn_available():
    return _has_sklearn


def is_scatter_available():
    return _scatter_available


DATASETS_IMPORT_ERROR = """
{0} requires the 🤗 Datasets library but it was not found in your environment. You can install it with:
```
pip install datasets
```
In a notebook or a colab, you can install it by executing a cell with
```
!pip install datasets
```
then restarting your kernel.

Note that if you have a local folder named `datasets` or a local python file named `datasets.py` in your current
working directory, python may try to import this instead of the 🤗 Datasets library. You should rename this folder or
that python file if that's the case.
"""


# docstyle-ignore
TOKENIZERS_IMPORT_ERROR = """
{0} requires the 🤗 Tokenizers library but it was not found in your environment. You can install it with:
```
pip install tokenizers
```
In a notebook or a colab, you can install it by executing a cell with
```
!pip install tokenizers
```
"""


# docstyle-ignore
SENTENCEPIECE_IMPORT_ERROR = """
{0} requires the SentencePiece library but it was not found in your environment. Checkout the instructions on the
installation page of its repo: https://github.com/google/sentencepiece#installation and follow the ones
that match your environment.
"""


# docstyle-ignore
FAISS_IMPORT_ERROR = """
{0} requires the faiss library but it was not found in your environment. Checkout the instructions on the
installation page of its repo: https://github.com/facebookresearch/faiss/blob/master/INSTALL.md and follow the ones
that match your environment.
"""


# docstyle-ignore
PYTORCH_IMPORT_ERROR = """
{0} requires the PyTorch library but it was not found in your environment. Checkout the instructions on the
installation page: https://pytorch.org/get-started/locally/ and follow the ones that match your environment.
"""


# docstyle-ignore
SKLEARN_IMPORT_ERROR = """
{0} requires the scikit-learn library but it was not found in your environment. You can install it with:
```
pip install -U scikit-learn
```
In a notebook or a colab, you can install it by executing a cell with
```
!pip install -U scikit-learn
```
"""


# docstyle-ignore
TENSORFLOW_IMPORT_ERROR = """
{0} requires the TensorFlow library but it was not found in your environment. Checkout the instructions on the
installation page: https://www.tensorflow.org/install and follow the ones that match your environment.
"""


# docstyle-ignore
FLAX_IMPORT_ERROR = """
{0} requires the FLAX library but it was not found in your environment. Checkout the instructions on the
installation page: https://github.com/google/flax and follow the ones that match your environment.
"""

SCATTER_IMPORT_ERROR = """
{0} requires the torch-scatter library but it was not found in your environment. You can install it with pip as
explained here: https://github.com/rusty1s/pytorch_scatter.

"""


def requires_datasets(obj):
    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    if not is_datasets_available():
        raise ImportError(DATASETS_IMPORT_ERROR.format(name))


def requires_faiss(obj):
    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    if not is_faiss_available():
        raise ImportError(FAISS_IMPORT_ERROR.format(name))


def requires_pytorch(obj):
    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    if not is_torch_available():
        raise ImportError(PYTORCH_IMPORT_ERROR.format(name))


def requires_sklearn(obj):
    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    if not is_sklearn_available():
        raise ImportError(SKLEARN_IMPORT_ERROR.format(name))


def requires_tf(obj):
    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    if not is_tf_available():
        raise ImportError(TENSORFLOW_IMPORT_ERROR.format(name))


def requires_flax(obj):
    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    if not is_flax_available():
        raise ImportError(FLAX_IMPORT_ERROR.format(name))


def requires_tokenizers(obj):
    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    if not is_tokenizers_available():
        raise ImportError(TOKENIZERS_IMPORT_ERROR.format(name))


def requires_sentencepiece(obj):
    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    if not is_sentencepiece_available():
        raise ImportError(SENTENCEPIECE_IMPORT_ERROR.format(name))


def requires_scatter(obj):
    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    if not is_scatter_available():
        raise ImportError(SCATTER_IMPORT_ERROR.format(name))


def add_start_docstrings(*docstr):
    def docstring_decorator(fn):
        fn.__doc__ = "".join(docstr) + (fn.__doc__ if fn.__doc__ is not None else "")
        return fn

    return docstring_decorator


def add_start_docstrings_to_model_forward(*docstr):
    def docstring_decorator(fn):
        class_name = ":class:`~transformers.{}`".format(fn.__qualname__.split(".")[0])
        intro = "   The {} forward method, overrides the :func:`__call__` special method.".format(class_name)
        note = r"""

    .. note::
        Although the recipe for forward pass needs to be defined within this function, one should call the
        :class:`Module` instance afterwards instead of this since the former takes care of running the pre and post
        processing steps while the latter silently ignores them.
        """
        fn.__doc__ = intro + note + "".join(docstr) + (fn.__doc__ if fn.__doc__ is not None else "")
        return fn

    return docstring_decorator


def add_end_docstrings(*docstr):
    def docstring_decorator(fn):
        fn.__doc__ = fn.__doc__ + "".join(docstr)
        return fn

    return docstring_decorator


PT_RETURN_INTRODUCTION = r"""
    Returns:
        :class:`~{full_output_type}` or :obj:`tuple(torch.FloatTensor)`: A :class:`~{full_output_type}` (if
        ``return_dict=True`` is passed or when ``config.return_dict=True``) or a tuple of :obj:`torch.FloatTensor`
        comprising various elements depending on the configuration (:class:`~transformers.{config_class}`) and inputs.

"""


TF_RETURN_INTRODUCTION = r"""
    Returns:
        :class:`~{full_output_type}` or :obj:`tuple(tf.Tensor)`: A :class:`~{full_output_type}` (if
        ``return_dict=True`` is passed or when ``config.return_dict=True``) or a tuple of :obj:`tf.Tensor` comprising
        various elements depending on the configuration (:class:`~transformers.{config_class}`) and inputs.

"""


def _get_indent(t):
    """Returns the indentation in the first line of t"""
    search = re.search(r"^(\s*)\S", t)
    return "" if search is None else search.groups()[0]


def _convert_output_args_doc(output_args_doc):
    """Convert output_args_doc to display properly."""
    # Split output_arg_doc in blocks argument/description
    indent = _get_indent(output_args_doc)
    blocks = []
    current_block = ""
    for line in output_args_doc.split("\n"):
        # If the indent is the same as the beginning, the line is the name of new arg.
        if _get_indent(line) == indent:
            if len(current_block) > 0:
                blocks.append(current_block[:-1])
            current_block = f"{line}\n"
        else:
            # Otherwise it's part of the description of the current arg.
            # We need to remove 2 spaces to the indentation.
            current_block += f"{line[2:]}\n"
    blocks.append(current_block[:-1])

    # Format each block for proper rendering
    for i in range(len(blocks)):
        blocks[i] = re.sub(r"^(\s+)(\S+)(\s+)", r"\1- **\2**\3", blocks[i])
        blocks[i] = re.sub(r":\s*\n\s*(\S)", r" -- \1", blocks[i])

    return "\n".join(blocks)


def _prepare_output_docstrings(output_type, config_class):
    """
    Prepares the return part of the docstring using `output_type`.
    """
    docstrings = output_type.__doc__

    # Remove the head of the docstring to keep the list of args only
    lines = docstrings.split("\n")
    i = 0
    while i < len(lines) and re.search(r"^\s*(Args|Parameters):\s*$", lines[i]) is None:
        i += 1
    if i < len(lines):
        docstrings = "\n".join(lines[(i + 1) :])
        docstrings = _convert_output_args_doc(docstrings)

    # Add the return introduction
    full_output_type = f"{output_type.__module__}.{output_type.__name__}"
    intro = TF_RETURN_INTRODUCTION if output_type.__name__.startswith("TF") else PT_RETURN_INTRODUCTION
    intro = intro.format(full_output_type=full_output_type, config_class=config_class)
    return intro + docstrings


PT_TOKEN_CLASSIFICATION_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import torch

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="pt")
        >>> labels = torch.tensor([1] * inputs["input_ids"].size(1)).unsqueeze(0)  # Batch size 1

        >>> outputs = model(**inputs, labels=labels)
        >>> loss = outputs.loss
        >>> logits = outputs.logits
"""

PT_QUESTION_ANSWERING_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import torch

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> question, text = "Who was Jim Henson?", "Jim Henson was a nice puppet"
        >>> inputs = tokenizer(question, text, return_tensors='pt')
        >>> start_positions = torch.tensor([1])
        >>> end_positions = torch.tensor([3])

        >>> outputs = model(**inputs, start_positions=start_positions, end_positions=end_positions)
        >>> loss = outputs.loss
        >>> start_scores = outputs.start_logits
        >>> end_scores = outputs.end_logits
"""

PT_SEQUENCE_CLASSIFICATION_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import torch

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="pt")
        >>> labels = torch.tensor([1]).unsqueeze(0)  # Batch size 1
        >>> outputs = model(**inputs, labels=labels)
        >>> loss = outputs.loss
        >>> logits = outputs.logits
"""

PT_MASKED_LM_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import torch

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("The capital of France is {mask}.", return_tensors="pt")
        >>> labels = tokenizer("The capital of France is Paris.", return_tensors="pt")["input_ids"]

        >>> outputs = model(**inputs, labels=labels)
        >>> loss = outputs.loss
        >>> logits = outputs.logits
"""

PT_BASE_MODEL_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import torch

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="pt")
        >>> outputs = model(**inputs)

        >>> last_hidden_states = outputs.last_hidden_state
"""

PT_MULTIPLE_CHOICE_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import torch

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> prompt = "In Italy, pizza served in formal settings, such as at a restaurant, is presented unsliced."
        >>> choice0 = "It is eaten with a fork and a knife."
        >>> choice1 = "It is eaten while held in the hand."
        >>> labels = torch.tensor(0).unsqueeze(0)  # choice0 is correct (according to Wikipedia ;)), batch size 1

        >>> encoding = tokenizer([[prompt, prompt], [choice0, choice1]], return_tensors='pt', padding=True)
        >>> outputs = model(**{{k: v.unsqueeze(0) for k,v in encoding.items()}}, labels=labels)  # batch size is 1

        >>> # the linear classifier still needs to be trained
        >>> loss = outputs.loss
        >>> logits = outputs.logits
"""

PT_CAUSAL_LM_SAMPLE = r"""
    Example::

        >>> import torch
        >>> from transformers import {tokenizer_class}, {model_class}

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="pt")
        >>> outputs = model(**inputs, labels=inputs["input_ids"])
        >>> loss = outputs.loss
        >>> logits = outputs.logits
"""

TF_TOKEN_CLASSIFICATION_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import tensorflow as tf

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="tf")
        >>> input_ids = inputs["input_ids"]
        >>> inputs["labels"] = tf.reshape(tf.constant([1] * tf.size(input_ids).numpy()), (-1, tf.size(input_ids))) # Batch size 1

        >>> outputs = model(inputs)
        >>> loss = outputs.loss
        >>> logits = outputs.logits
"""

TF_QUESTION_ANSWERING_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import tensorflow as tf

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> question, text = "Who was Jim Henson?", "Jim Henson was a nice puppet"
        >>> input_dict = tokenizer(question, text, return_tensors='tf')
        >>> outputs = model(input_dict)
        >>> start_logits = outputs.start_logits
        >>> end_logits = outputs.end_logits

        >>> all_tokens = tokenizer.convert_ids_to_tokens(input_dict["input_ids"].numpy()[0])
        >>> answer = ' '.join(all_tokens[tf.math.argmax(start_logits, 1)[0] : tf.math.argmax(end_logits, 1)[0]+1])
"""

TF_SEQUENCE_CLASSIFICATION_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import tensorflow as tf

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="tf")
        >>> inputs["labels"] = tf.reshape(tf.constant(1), (-1, 1)) # Batch size 1

        >>> outputs = model(inputs)
        >>> loss = outputs.loss
        >>> logits = outputs.logits
"""

TF_MASKED_LM_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import tensorflow as tf

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("The capital of France is {mask}.", return_tensors="tf")
        >>> inputs["labels"] = tokenizer("The capital of France is Paris.", return_tensors="tf")["input_ids"]

        >>> outputs = model(inputs)
        >>> loss = outputs.loss
        >>> logits = outputs.logits
"""

TF_BASE_MODEL_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import tensorflow as tf

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="tf")
        >>> outputs = model(inputs)

        >>> last_hidden_states = outputs.last_hidden_states
"""

TF_MULTIPLE_CHOICE_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import tensorflow as tf

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> prompt = "In Italy, pizza served in formal settings, such as at a restaurant, is presented unsliced."
        >>> choice0 = "It is eaten with a fork and a knife."
        >>> choice1 = "It is eaten while held in the hand."

        >>> encoding = tokenizer([[prompt, prompt], [choice0, choice1]], return_tensors='tf', padding=True)
        >>> inputs = {{k: tf.expand_dims(v, 0) for k, v in encoding.items()}}
        >>> outputs = model(inputs)  # batch size is 1

        >>> # the linear classifier still needs to be trained
        >>> logits = outputs.logits
"""

TF_CAUSAL_LM_SAMPLE = r"""
    Example::

        >>> from transformers import {tokenizer_class}, {model_class}
        >>> import tensorflow as tf

        >>> tokenizer = {tokenizer_class}.from_pretrained('{checkpoint}')
        >>> model = {model_class}.from_pretrained('{checkpoint}', return_dict=True)

        >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="tf")
        >>> outputs = model(inputs)
        >>> logits = outputs.logits
"""


def add_code_sample_docstrings(
    *docstr, tokenizer_class=None, checkpoint=None, output_type=None, config_class=None, mask=None
):
    def docstring_decorator(fn):
        model_class = fn.__qualname__.split(".")[0]
        is_tf_class = model_class[:2] == "TF"
        doc_kwargs = dict(model_class=model_class, tokenizer_class=tokenizer_class, checkpoint=checkpoint)

        if "SequenceClassification" in model_class:
            code_sample = TF_SEQUENCE_CLASSIFICATION_SAMPLE if is_tf_class else PT_SEQUENCE_CLASSIFICATION_SAMPLE
        elif "QuestionAnswering" in model_class:
            code_sample = TF_QUESTION_ANSWERING_SAMPLE if is_tf_class else PT_QUESTION_ANSWERING_SAMPLE
        elif "TokenClassification" in model_class:
            code_sample = TF_TOKEN_CLASSIFICATION_SAMPLE if is_tf_class else PT_TOKEN_CLASSIFICATION_SAMPLE
        elif "MultipleChoice" in model_class:
            code_sample = TF_MULTIPLE_CHOICE_SAMPLE if is_tf_class else PT_MULTIPLE_CHOICE_SAMPLE
        elif "MaskedLM" in model_class or model_class in ["FlaubertWithLMHeadModel", "XLMWithLMHeadModel"]:
            doc_kwargs["mask"] = "[MASK]" if mask is None else mask
            code_sample = TF_MASKED_LM_SAMPLE if is_tf_class else PT_MASKED_LM_SAMPLE
        elif "LMHead" in model_class:
            code_sample = TF_CAUSAL_LM_SAMPLE if is_tf_class else PT_CAUSAL_LM_SAMPLE
        elif "Model" in model_class or "Encoder" in model_class:
            code_sample = TF_BASE_MODEL_SAMPLE if is_tf_class else PT_BASE_MODEL_SAMPLE
        else:
            raise ValueError(f"Docstring can't be built for model {model_class}")

        output_doc = _prepare_output_docstrings(output_type, config_class) if output_type is not None else ""
        built_doc = code_sample.format(**doc_kwargs)
        fn.__doc__ = (fn.__doc__ or "") + "".join(docstr) + output_doc + built_doc
        return fn

    return docstring_decorator


def replace_return_docstrings(output_type=None, config_class=None):
    def docstring_decorator(fn):
        docstrings = fn.__doc__
        lines = docstrings.split("\n")
        i = 0
        while i < len(lines) and re.search(r"^\s*Returns?:\s*$", lines[i]) is None:
            i += 1
        if i < len(lines):
            lines[i] = _prepare_output_docstrings(output_type, config_class)
            docstrings = "\n".join(lines)
        else:
            raise ValueError(
                f"The function {fn} should have an empty 'Return:' or 'Returns:' in its docstring as placeholder, current docstring is:\n{docstrings}"
            )
        fn.__doc__ = docstrings
        return fn

    return docstring_decorator


def is_remote_url(url_or_filename):
    parsed = urlparse(url_or_filename)
    return parsed.scheme in ("http", "https")


def hf_bucket_url(model_id: str, filename: str, use_cdn=True, mirror=None) -> str:
    """
    Resolve a model identifier, and a file name, to a HF-hosted url on either S3 or Cloudfront (a Content Delivery
    Network, or CDN).

    Cloudfront is replicated over the globe so downloads are way faster for the end user (and it also lowers our
    bandwidth costs). However, it is more aggressively cached by default, so may not always reflect the latest changes
    to the underlying file (default TTL is 24 hours).

    In terms of client-side caching from this library, even though Cloudfront relays the ETags from S3, using one or
    the other (or switching from one to the other) will affect caching: cached files are not shared between the two
    because the cached file's name contains a hash of the url.
    """
    endpoint = (
        PRESET_MIRROR_DICT.get(mirror, mirror)
        if mirror
        else CLOUDFRONT_DISTRIB_PREFIX
        if use_cdn
        else S3_BUCKET_PREFIX
    )
    legacy_format = "/" not in model_id
    if legacy_format:
        return f"{endpoint}/{model_id}-{filename}"
    else:
        return f"{endpoint}/{model_id}/{filename}"


def url_to_filename(url, etag=None):
    """
    Convert `url` into a hashed filename in a repeatable way. If `etag` is specified, append its hash to the url's,
    delimited by a period. If the url ends with .h5 (Keras HDF5 weights) adds '.h5' to the name so that TF 2.0 can
    identify it as a HDF5 file (see
    https://github.com/tensorflow/tensorflow/blob/00fad90125b18b80fe054de1055770cfb8fe4ba3/tensorflow/python/keras/engine/network.py#L1380)
    """
    url_bytes = url.encode("utf-8")
    url_hash = sha256(url_bytes)
    filename = url_hash.hexdigest()

    if etag:
        etag_bytes = etag.encode("utf-8")
        etag_hash = sha256(etag_bytes)
        filename += "." + etag_hash.hexdigest()

    if url.endswith(".h5"):
        filename += ".h5"

    return filename


def filename_to_url(filename, cache_dir=None):
    """
    Return the url and etag (which may be ``None``) stored for `filename`. Raise ``EnvironmentError`` if `filename` or
    its stored metadata do not exist.
    """
    if cache_dir is None:
        cache_dir = TRANSFORMERS_CACHE
    if isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)

    cache_path = os.path.join(cache_dir, filename)
    if not os.path.exists(cache_path):
        raise EnvironmentError("file {} not found".format(cache_path))

    meta_path = cache_path + ".json"
    if not os.path.exists(meta_path):
        raise EnvironmentError("file {} not found".format(meta_path))

    with open(meta_path, encoding="utf-8") as meta_file:
        metadata = json.load(meta_file)
    url = metadata["url"]
    etag = metadata["etag"]

    return url, etag


def cached_path(
    url_or_filename,
    cache_dir=None,
    force_download=False,
    proxies=None,
    resume_download=False,
    user_agent: Union[Dict, str, None] = None,
    extract_compressed_file=False,
    force_extract=False,
    local_files_only=False,
) -> Optional[str]:
    """
    Given something that might be a URL (or might be a local path), determine which. If it's a URL, download the file
    and cache it, and return the path to the cached file. If it's already a local path, make sure the file exists and
    then return the path

    Args:
        cache_dir: specify a cache directory to save the file to (overwrite the default cache dir).
        force_download: if True, re-download the file even if it's already cached in the cache dir.
        resume_download: if True, resume the download if incompletely received file is found.
        user_agent: Optional string or dict that will be appended to the user-agent on remote requests.
        extract_compressed_file: if True and the path point to a zip or tar file, extract the compressed
            file in a folder along the archive.
        force_extract: if True when extract_compressed_file is True and the archive was already extracted,
            re-extract the archive and override the folder where it was extracted.

    Return:
        None in case of non-recoverable file (non-existent or inaccessible url + no cache on disk). Local path (string)
        otherwise
    """
    if cache_dir is None:
        cache_dir = TRANSFORMERS_CACHE
    if isinstance(url_or_filename, Path):
        url_or_filename = str(url_or_filename)
    if isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)

    if is_remote_url(url_or_filename):
        # URL, so get it from the cache (downloading if necessary)
        output_path = get_from_cache(
            url_or_filename,
            cache_dir=cache_dir,
            force_download=force_download,
            proxies=proxies,
            resume_download=resume_download,
            user_agent=user_agent,
            local_files_only=local_files_only,
        )
    elif os.path.exists(url_or_filename):
        # File, and it exists.
        output_path = url_or_filename
    elif urlparse(url_or_filename).scheme == "":
        # File, but it doesn't exist.
        raise EnvironmentError("file {} not found".format(url_or_filename))
    else:
        # Something unknown
        raise ValueError("unable to parse {} as a URL or as a local path".format(url_or_filename))

    if extract_compressed_file:
        if not is_zipfile(output_path) and not tarfile.is_tarfile(output_path):
            return output_path

        # Path where we extract compressed archives
        # We avoid '.' in dir name and add "-extracted" at the end: "./model.zip" => "./model-zip-extracted/"
        output_dir, output_file = os.path.split(output_path)
        output_extract_dir_name = output_file.replace(".", "-") + "-extracted"
        output_path_extracted = os.path.join(output_dir, output_extract_dir_name)

        if os.path.isdir(output_path_extracted) and os.listdir(output_path_extracted) and not force_extract:
            return output_path_extracted

        # Prevent parallel extractions
        lock_path = output_path + ".lock"
        with FileLock(lock_path):
            shutil.rmtree(output_path_extracted, ignore_errors=True)
            os.makedirs(output_path_extracted)
            if is_zipfile(output_path):
                with ZipFile(output_path, "r") as zip_file:
                    zip_file.extractall(output_path_extracted)
                    zip_file.close()
            elif tarfile.is_tarfile(output_path):
                tar_file = tarfile.open(output_path)
                tar_file.extractall(output_path_extracted)
                tar_file.close()
            else:
                raise EnvironmentError("Archive format of {} could not be identified".format(output_path))

        return output_path_extracted

    return output_path


def http_get(url, temp_file, proxies=None, resume_size=0, user_agent: Union[Dict, str, None] = None):
    ua = "transformers/{}; python/{}".format(__version__, sys.version.split()[0])
    if is_torch_available():
        ua += "; torch/{}".format(torch.__version__)
    if is_tf_available():
        ua += "; tensorflow/{}".format(tf.__version__)
    if isinstance(user_agent, dict):
        ua += "; " + "; ".join("{}/{}".format(k, v) for k, v in user_agent.items())
    elif isinstance(user_agent, str):
        ua += "; " + user_agent
    headers = {"user-agent": ua}
    if resume_size > 0:
        headers["Range"] = "bytes=%d-" % (resume_size,)
    response = requests.get(url, stream=True, proxies=proxies, headers=headers)
    if response.status_code == 416:  # Range not satisfiable
        return
    content_length = response.headers.get("Content-Length")
    total = resume_size + int(content_length) if content_length is not None else None
    progress = tqdm(
        unit="B",
        unit_scale=True,
        total=total,
        initial=resume_size,
        desc="Downloading",
        disable=bool(logging.get_verbosity() == logging.NOTSET),
    )
    for chunk in response.iter_content(chunk_size=1024):
        if chunk:  # filter out keep-alive new chunks
            progress.update(len(chunk))
            temp_file.write(chunk)
    progress.close()


def get_from_cache(
    url,
    cache_dir=None,
    force_download=False,
    proxies=None,
    etag_timeout=10,
    resume_download=False,
    user_agent: Union[Dict, str, None] = None,
    local_files_only=False,
) -> Optional[str]:
    """
    Given a URL, look for the corresponding file in the local cache. If it's not there, download it. Then return the
    path to the cached file.

    Return:
        None in case of non-recoverable file (non-existent or inaccessible url + no cache on disk). Local path (string)
        otherwise
    """
    if cache_dir is None:
        cache_dir = TRANSFORMERS_CACHE
    if isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)

    os.makedirs(cache_dir, exist_ok=True)

    etag = None
    if not local_files_only:
        try:
            response = requests.head(url, allow_redirects=True, proxies=proxies, timeout=etag_timeout)
            if response.status_code == 200:
                etag = response.headers.get("ETag")
        except (EnvironmentError, requests.exceptions.Timeout):
            # etag is already None
            pass

    filename = url_to_filename(url, etag)

    # get cache path to put the file
    cache_path = os.path.join(cache_dir, filename)

    # etag is None = we don't have a connection, or url doesn't exist, or is otherwise inaccessible.
    # try to get the last downloaded one
    if etag is None:
        if os.path.exists(cache_path):
            return cache_path
        else:
            matching_files = [
                file
                for file in fnmatch.filter(os.listdir(cache_dir), filename.split(".")[0] + ".*")
                if not file.endswith(".json") and not file.endswith(".lock")
            ]
            if len(matching_files) > 0:
                return os.path.join(cache_dir, matching_files[-1])
            else:
                # If files cannot be found and local_files_only=True,
                # the models might've been found if local_files_only=False
                # Notify the user about that
                if local_files_only:
                    raise ValueError(
                        "Cannot find the requested files in the cached path and outgoing traffic has been"
                        " disabled. To enable model look-ups and downloads online, set 'local_files_only'"
                        " to False."
                    )
                return None

    # From now on, etag is not None.
    if os.path.exists(cache_path) and not force_download:
        return cache_path

    # Prevent parallel downloads of the same file with a lock.
    lock_path = cache_path + ".lock"
    with FileLock(lock_path):

        # If the download just completed while the lock was activated.
        if os.path.exists(cache_path) and not force_download:
            # Even if returning early like here, the lock will be released.
            return cache_path

        if resume_download:
            incomplete_path = cache_path + ".incomplete"

            @contextmanager
            def _resumable_file_manager():
                with open(incomplete_path, "a+b") as f:
                    yield f

            temp_file_manager = _resumable_file_manager
            if os.path.exists(incomplete_path):
                resume_size = os.stat(incomplete_path).st_size
            else:
                resume_size = 0
        else:
            temp_file_manager = partial(tempfile.NamedTemporaryFile, dir=cache_dir, delete=False)
            resume_size = 0

        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        with temp_file_manager() as temp_file:
            logger.info("%s not found in cache or force_download set to True, downloading to %s", url, temp_file.name)

            http_get(url, temp_file, proxies=proxies, resume_size=resume_size, user_agent=user_agent)

        logger.info("storing %s in cache at %s", url, cache_path)
        os.replace(temp_file.name, cache_path)

        logger.info("creating metadata file for %s", cache_path)
        meta = {"url": url, "etag": etag}
        meta_path = cache_path + ".json"
        with open(meta_path, "w") as meta_file:
            json.dump(meta, meta_file)

    return cache_path


class cached_property(property):
    """
    Descriptor that mimics @property but caches output in member variable.

    From tensorflow_datasets

    Built-in in functools from Python 3.8.
    """

    def __get__(self, obj, objtype=None):
        # See docs.python.org/3/howto/descriptor.html#properties
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        attr = "__cached_" + self.fget.__name__
        cached = getattr(obj, attr, None)
        if cached is None:
            cached = self.fget(obj)
            setattr(obj, attr, cached)
        return cached


def torch_required(func):
    # Chose a different decorator name than in tests so it's clear they are not the same.
    @wraps(func)
    def wrapper(*args, **kwargs):
        if is_torch_available():
            return func(*args, **kwargs)
        else:
            raise ImportError(f"Method `{func.__name__}` requires PyTorch.")

    return wrapper


def tf_required(func):
    # Chose a different decorator name than in tests so it's clear they are not the same.
    @wraps(func)
    def wrapper(*args, **kwargs):
        if is_tf_available():
            return func(*args, **kwargs)
        else:
            raise ImportError(f"Method `{func.__name__}` requires TF.")

    return wrapper


def is_tensor(x):
    """ Tests if ``x`` is a :obj:`torch.Tensor`, :obj:`tf.Tensor` or :obj:`np.ndarray`. """
    if is_torch_available():
        import torch

        if isinstance(x, torch.Tensor):
            return True
    if is_tf_available():
        import tensorflow as tf

        if isinstance(x, tf.Tensor):
            return True
    return isinstance(x, np.ndarray)


class ModelOutput(OrderedDict):
    """
    Base class for all model outputs as dataclass. Has a ``__getitem__`` that allows indexing by integer or slice (like
    a tuple) or strings (like a dictionary) that will ignore the ``None`` attributes. Otherwise behaves like a regular
    python dictionary.

    .. warning::
        You can't unpack a :obj:`ModelOutput` directly. Use the :meth:`~transformers.file_utils.ModelOutput.to_tuple`
        method to convert it to a tuple before.
    """

    def __post_init__(self):
        class_fields = fields(self)

        # Safety and consistency checks
        assert len(class_fields), f"{self.__class__.__name__} has no fields."
        assert all(
            field.default is None for field in class_fields[1:]
        ), f"{self.__class__.__name__} should not have more than one required field."

        first_field = getattr(self, class_fields[0].name)
        other_fields_are_none = all(getattr(self, field.name) is None for field in class_fields[1:])

        if other_fields_are_none and not is_tensor(first_field):
            try:
                iterator = iter(first_field)
                first_field_iterator = True
            except TypeError:
                first_field_iterator = False

            # if we provided an iterator as first field and the iterator is a (key, value) iterator
            # set the associated fields
            if first_field_iterator:
                for element in iterator:
                    if (
                        not isinstance(element, (list, tuple))
                        or not len(element) == 2
                        or not isinstance(element[0], str)
                    ):
                        break
                    setattr(self, element[0], element[1])
                    if element[1] is not None:
                        self[element[0]] = element[1]
            elif first_field is not None:
                self[class_fields[0].name] = first_field
        else:
            for field in class_fields:
                v = getattr(self, field.name)
                if v is not None:
                    self[field.name] = v

    def __delitem__(self, *args, **kwargs):
        raise Exception(f"You cannot use ``__delitem__`` on a {self.__class__.__name__} instance.")

    def setdefault(self, *args, **kwargs):
        raise Exception(f"You cannot use ``setdefault`` on a {self.__class__.__name__} instance.")

    def pop(self, *args, **kwargs):
        raise Exception(f"You cannot use ``pop`` on a {self.__class__.__name__} instance.")

    def update(self, *args, **kwargs):
        raise Exception(f"You cannot use ``update`` on a {self.__class__.__name__} instance.")

    def __getitem__(self, k):
        if isinstance(k, str):
            inner_dict = {k: v for (k, v) in self.items()}
            return inner_dict[k]
        else:
            return self.to_tuple()[k]

    def __setattr__(self, name, value):
        if name in self.keys() and value is not None:
            # Don't call self.__setitem__ to avoid recursion errors
            super().__setitem__(name, value)
        super().__setattr__(name, value)

    def __setitem__(self, key, value):
        # Will raise a KeyException if needed
        super().__setitem__(key, value)
        # Don't call self.__setattr__ to avoid recursion errors
        super().__setattr__(key, value)

    def to_tuple(self) -> Tuple[Any]:
        """
        Convert self to a tuple containing all the attributes/keys that are not ``None``.
        """
        return tuple(self[k] for k in self.keys())



import json
import gzip
import datasets
import os
from PIL import Image
ROOT_DIR = os.path.dirname(__file__)
_CITATION = """\
@article{talmor2021multimodalqa,
  title={MultiModalQA: Complex Question Answering over Text, Tables and Images},
  author={Talmor, Alon and Yoran, Ori and Catav, Amnon and Lahav, Dan and Wang, Yizhong and Asai, Akari and Ilharco, Gabriel and Hajishirzi, Hannaneh and Berant, Jonathan},
  journal={arXiv preprint arXiv:2104.06039},
  year={2021}
}
"""

_DESCRIPTION = """\
This dataset is obtained from the official release of the MMQA.
"""

_HOMEPAGE = "https://github.com/allenai/multimodalqa"

_LICENSE = "MIT License"

_TRAINING_FILE = "MMQA_train.jsonl.gz"
_DEV_FILE = "MMQA_dev.jsonl.gz"
_TEST_FILE = "MMQA_test.jsonl.gz"
_TEXTS_FILE = "MMQA_texts.jsonl.gz"
_TABLES_FILE = "MMQA_tables.jsonl.gz"
_PASSAGE_FILE = "MMQA_texts.jsonl.gz"
_IMAGES_INFO_FILE = "MMQA_images.jsonl.gz"
_IMAGES_FILE = "final_dataset_images"
_DATA_PATH = "/home/lfy/lwh/data/mmqa"
# _DATA_PATH = ROOT_DIR
def test():
    with open(_DATA_PATH + "/MMQA_images.jsonl.gz", "r") as f:
        images_info = [json.loads(_line) for _line in f.readlines()]
        print(images_info[0])


class Images(object): # 存所有的图片信息
    def __init__(self, images_info_path, pictures_path):
        self.images_info_path = images_info_path
        self.pictures_path = pictures_path

        self.images_info_dict = {}
        self.images_pic_dict = {}

        with gzip.open(images_info_path, "r") as f:
            images_info = [json.loads(_line) for _line in f.readlines()]
        for image_info in images_info:
            # {
            #   "title": "Nintendo Entertainment System",
            #   "url": "https://en.wikipedia.org/wiki/Nintendo_Entertainment_System",
            #   "id": "991237d4689fa65507e9528a1fab7de3",
            #   "path": "991237d4689fa65507e9528a1fab7de3.jpg"
            # }
            self.images_info_dict[image_info['id']] = image_info

    def load_image(self, pic_id, open_by_pillow=False):
        picture_absolute_path = os.path.join(self.pictures_path, self.images_info_dict[pic_id]['path'])
        if open_by_pillow:
            return Image.open(picture_absolute_path)
        else:
            return picture_absolute_path

    def __getitem__(self, id):
        return {"pic": self.load_image(id), **self.images_info_dict[id]}


class MMQA(datasets.GeneratorBasedBuilder):
    """The MMQA dataset"""

    def _info(self):
        return datasets.DatasetInfo( # 类似于定义了一个数据库, 也相当于定义了返回的字典的结构
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "id": datasets.Value("string"),
                    "question": datasets.Value("string"),
                    "table": datasets.features.Sequence(
                        {
                            "table_id": datasets.Value("string"),
                            "title": datasets.Value("string"),
                            "caption": datasets.Value("string"),
                            "header": datasets.features.Sequence(datasets.Value("string")),
                            "rows": datasets.features.Sequence(datasets.features.Sequence(datasets.Value("string"))),
                            "rows_with_links": datasets.features.Sequence(datasets.features.Sequence(
                                datasets.features.Sequence(datasets.features.Sequence(
                                    datasets.Value("string")))))
                        }
                    ),
                    "images": datasets.features.Sequence(
                        {
                            "id": datasets.Value("string"),
                            "title": datasets.Value("string"),
                            "url": datasets.Value("string"),
                            "path": datasets.Value("string"),
                            "pic": datasets.Value("string")
                        }
                    ),
                    "passages": datasets.features.Sequence(
                        {
                            "id": datasets.Value("string"),
                            "title": datasets.Value("string"),
                            "url": datasets.Value("string"),
                            "text": datasets.Value("string")
                        }
                    ),
                    "answer_text": datasets.Value("string"),
                    "supporting_context": datasets.features.Sequence(
                        {
                            "doc_id": datasets.Value("string"),
                            "doc_part": datasets.Value("string")
                        }
                    ),
                    "type": datasets.Value("string")
                }
            ),
            supervised_keys=None,
            homepage=_HOMEPAGE,
            license=_LICENSE,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager=None):
        """Returns SplitGenerators."""
        local_files = {
            "train": os.path.join(_DATA_PATH, _TRAINING_FILE),
            "dev": os.path.join(_DATA_PATH, _DEV_FILE),
            "test": os.path.join(_DATA_PATH, _TEST_FILE),
            "texts": os.path.join(_DATA_PATH, _TEXTS_FILE),
            "tables": os.path.join(_DATA_PATH, _TABLES_FILE),
            "passages": os.path.join(_DATA_PATH, _PASSAGE_FILE),
            "images_info": os.path.join(_DATA_PATH, _IMAGES_INFO_FILE),
            "images": os.path.join(_DATA_PATH, _IMAGES_FILE)
        }
        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={
                    "split": "train",
                    "file_path": local_files["train"],
                    "table_path": local_files["tables"],
                    "images_path": os.path.join(local_files["images"], _IMAGES_FILE),
                    "images_info_path": local_files["images_info"],
                    "passage_path": local_files["passages"]
                }),
            datasets.SplitGenerator(
                name=datasets.Split.VALIDATION,
                gen_kwargs={
                    "split": "validation",
                    "file_path": local_files["dev"],
                    "table_path": local_files["tables"],
                    "images_path": os.path.join(local_files["images"], _IMAGES_FILE),
                    "images_info_path": local_files["images_info"],
                    "passage_path": local_files["passages"]
                }),
            datasets.SplitGenerator(
                name=datasets.Split.TEST,
                gen_kwargs={
                    "split": "test",
                    "file_path": local_files["test"],
                    "table_path": local_files["tables"],
                    "images_path": os.path.join(local_files["images"], _IMAGES_FILE),
                    "images_info_path": local_files["images_info"],
                    "passage_path": local_files["passages"]
                }),
        ]

    def _generate_examples(self, split, file_path, table_path, images_path, images_info_path, passage_path):
        """Yields examples."""
        tables = {}
        with gzip.open(table_path, 'r') as f: # 表格
            for line in f:
                table = json.loads(line)
                tables[table["id"]] = table
        texts = {}
        with gzip.open(passage_path, 'r') as f: # 文章，段落
            for line in f:
                text = json.loads(line)
                texts[text["id"]] = text
        
        mmqa_images = Images(images_info_path, images_path) # 图片

        with gzip.open(file_path, 'r') as f: # 问题
            count = 0
            for idx, line in enumerate(f):
                example = json.loads(line)
                count += 1
                example_table = []
                example_images = []
                example_texts = []

                # load table
                table_id = example['metadata']['table_id']
                rows_with_links = []
                for row in tables[table_id]["table"]["table_rows"]:
                    rows_with_links.append([])
                    for cell in row:
                        text, title, url = [], [], []
                        for link in cell['links']:
                            text.append(link['text'])
                            title.append(link['wiki_title'])
                            url.append(link['url'])
                        rows_with_links[-1].append([text, title, url])

                example_table.append({
                    "table_id": table_id,
                    "title": tables[table_id]["title"],
                    "caption": tables[table_id]["table"]["table_name"],
                    "header": [column["column_name"] for column in tables[table_id]["table"]["header"]],
                    "rows": [[cell["text"] for cell in row] for row in tables[table_id]["table"]["table_rows"]],
                    "rows_with_links": rows_with_links
                })

                # load image_docs
                for image_doc_id in example['metadata']['image_doc_ids']:
                    example_images.append(mmqa_images[image_doc_id])

                # load text_docs
                for text_doc_id in example['metadata']['text_doc_ids']:
                    # {
                    #   "title": "Hillaryland",
                    #   "url": "https://en.wikipedia.org/wiki/Hillaryland",
                    #   "id": "a7d9e6350bafc46b700e4d0739a39594",
                    #   "text": "Hillaryland was the self-designated name of a group of core advisors to Hillary Clinton, when she was First Lady of the United States and again when, as United States Senator, she was one of the Democratic Party candidates for President in the 2008 U.S. election."
                    # }
                    example_texts.append(texts[text_doc_id])

                if split in ['train', 'validation']:
                    yield count, {
                        "id": example["qid"],
                        "question": example["question"],
                        "table": example_table,
                        "images": example_images,
                        "passages": example_texts,
                        "answer_text": " | ".join([str(answer["answer"]) for answer in example["answers"]]),
                        "supporting_context": example['supporting_context'],
                        "type": example['metadata']['type']
                        #    supporting_context be like
                        #     {
                        #         "doc_id": "dcd7cb8f23737c6f38519c3770a6606f",
                        #         "doc_part": "table"
                        #     }
                    }
                else:
                    yield count, {
                        "id": example["qid"],
                        "question": example["question"],
                        "table": example_table,
                        "images": example_images,
                        "passages": example_texts,
                        "answer_text": "",
                        "supporting_context": [],
                        "type": ""
                    }



import json
from abc import ABC, abstractmethod
from datasets import Dataset
from transformers import PreTrainedTokenizer, Seq2SeqTrainingArguments

from .template import Template
from .utils import DataTrainingArguments, IGNORE_INDEX


def preprocess_data(
        prompt_template: Template,
        dataset: Dataset,
        tokenizer: PreTrainedTokenizer,
        data_args: DataTrainingArguments,
        training_args: Seq2SeqTrainingArguments
) -> Dataset:
    column_names = list(dataset.column_names)

    # support question with a single answer or multiple answers
    def get_dialog(examples):
        for i in range(len(examples["prompt"])):
            if examples["prompt"][i] and examples["response"][i]:
                query, answer = examples["prompt"][i], examples["response"][i]
                query = query + "\n" + examples["query"][i] if examples["query"][i] else query
                prefix = examples["prefix"][i] if examples["prefix"][i] else ""
                dialog = prompt_template.get_dialog(query, answer, examples["history"][i], prefix)
                yield dialog

    def preprocess_supervised_dataset(examples):
        # build inputs with format `<bos> X Y <eos>` and labels with format `<ignore> ... <ignore> Y <eos>`
        # for input with history, we build multiple input-label pairs just like:
        # https://github.com/lm-sys/FastChat/blob/f17c092f64840fa6354ed52789dccb2daa793d0b/fastchat/train/train.py#L112
        model_inputs = {"input_ids": [], "labels": []}
        for dialog in get_dialog(examples):
            input_ids, labels = [], []

            for i in range(len(dialog) // 2):
                source_ids = tokenizer.encode(text=dialog[2 * i], add_special_tokens=False)
                target_ids = tokenizer.encode(text=dialog[2 * i + 1], add_special_tokens=False)

                if len(source_ids) > data_args.max_source_length - 1:  # bos token
                    source_ids = source_ids[:data_args.max_source_length - 1]
                if len(target_ids) > data_args.max_target_length - 1:  # eos token
                    target_ids = target_ids[:data_args.max_target_length - 1]

                input_ids += [tokenizer.bos_token_id] + source_ids + target_ids + [tokenizer.eos_token_id]
                labels += [IGNORE_INDEX] * (len(source_ids) + 1) + target_ids + [tokenizer.eos_token_id]

            if len(input_ids) > data_args.max_source_length + data_args.max_target_length:
                input_ids = input_ids[:data_args.max_source_length + data_args.max_target_length]
            if len(labels) > data_args.max_source_length + data_args.max_target_length:
                labels = labels[:data_args.max_source_length + data_args.max_target_length]

            model_inputs["input_ids"].append(input_ids)
            model_inputs["labels"].append(labels)
        return model_inputs

    def print_supervised_dataset_example(example):
        print("input_ids:\n{}".format(example["input_ids"]))
        print("inputs:\n{}".format(tokenizer.decode(example["input_ids"], skip_special_tokens=False)))
        print("label_ids:\n{}".format(example["labels"]))
        print("labels:\n{}".format(
            tokenizer.decode([d if d != IGNORE_INDEX else tokenizer.pad_token_id for d in example["labels"]],
                             skip_special_tokens=False)
        ))

    preprocess_function = preprocess_supervised_dataset

    with training_args.main_process_first(desc="dataset map pre-processing"):
        dataset = dataset.map(
            preprocess_function,
            batched=True,
            num_proc=data_args.preprocessing_num_workers,
            remove_columns=column_names,
            load_from_cache_file=not data_args.overwrite_cache,
            desc="Running tokenizer on dataset"
        )
        print_supervised_dataset_example(dataset[0])

        return dataset


class LLaMaDataset(ABC):
    def __init__(self, data_path: str):
        self.data_path = data_path
        self.dataset = self.__read_data_to_huggingface_dataset__(data_path)

    @abstractmethod
    def __read_data_to_huggingface_dataset__(self, data_path: str) -> Dataset:
        """
        Reading the data and preprocessing to the huggingface dataset with the column_names bellow:
        column_names = ["prefix", "prompt", "query", "response", "history"]
        :return: dataset: the huggingface Dataset
        """
        pass


class MultiHiertt(LLaMaDataset):
    def __read_data_to_huggingface_dataset__(self, data_path: str) -> Dataset:
        column_names = ["prefix", "prompt", "query", "response", "history"]
        data = json.load(
            open(data_path, 'r', encoding='utf-8')
        )
        dataset = []
        for one in data:
            query = ""
            for text_evidence in one["qa"]["text_evidence"]:
                query += (one["paragraphs"][text_evidence] + "\n")

            for table_evidence in one["qa"]["table_evidence"]:
                query += (one["table_description"][table_evidence] + "\n")
            dataset.append({
                "prefix": None,
                "prompt": "According to the information, use the operation in "
                          "[add, subtract, multiply, divide, exp, greater, table_sum, table_average, table_max, table_min] "
                          "to construct a program to answer the question:" + one["qa"]["question"],
                "query": query,
                "response": one["qa"]["program"],
                "history": None
            })

        huggingface_data = {column_name: [] for column_name in column_names}

        for data_sample in dataset:
            for column_name in column_names:
                huggingface_data[column_name].append(data_sample[column_name])

        hf_dataset = Dataset.from_dict(huggingface_data)
        return hf_dataset


class SKG(LLaMaDataset):
    def __read_data_to_huggingface_dataset__(self, data_path: str) -> Dataset:
        column_names = ["prefix", "prompt", "query", "response", "history"]
        data = json.load(
            open(data_path, 'r', encoding='utf-8')
        )
        dataset = []
        for one in data:
            dataset.append({
                "prefix": None,
                "prompt": "According to the table, try your best to answer the question: " + one["text_in"],
                "query": one["struct_in"],
                "response": one["seq_out"],
                "history": None
            })

        huggingface_data = {column_name: [] for column_name in column_names}

        for data_sample in dataset:
            for column_name in column_names:
                huggingface_data[column_name].append(data_sample[column_name])

        hf_dataset = Dataset.from_dict(huggingface_data)
        return hf_dataset


class CompAQT(LLaMaDataset):
    def __read_data_to_huggingface_dataset__(self, data_path: str) -> Dataset:
        column_names = ["prefix", "prompt", "query", "response", "history"]
        data = json.load(
            open(data_path, 'r', encoding='utf-8')
        )
        dataset = []
        for one in data:
            texts = "paragraphs: \n"
            tables = "table descriptions: \n"
            for k, v in one["qa"]["gold_inds"].items():
                if "text" in k:
                    texts += (v + "\n")
                if "table" in k:
                    tables += (v + "\n")
            dataset.append({
                "prefix": None,
                "prompt": "According to the paragraphs and table descriptions, try your best to answer the question: " + one["qa"]["question"],
                "query": texts + tables,
                "response": one["qa"]["program"],
                "history": None
            })

        huggingface_data = {column_name: [] for column_name in column_names}

        for data_sample in dataset:
            for column_name in column_names:
                huggingface_data[column_name].append(data_sample[column_name])

        hf_dataset = Dataset.from_dict(huggingface_data)
        return hf_dataset


from typing import List

from datasets import concatenate_datasets

from .data_collator import DynamicDataCollatorWithPadding
from .dataset import (
    LLaMaDataset, preprocess_data,

)
from .model import load_pretrained, plot_loss
from .peft_trainer import LogCallback
from .seq2seq import Seq2SeqPeftTrainer, ComputeMetrics
from .template import Template
from .utils import prepare_args, get_logits_processor


class LLaMaTrainer:
    def __init__(self, dataset_list: List[LLaMaDataset], template: Template):
        self.model_args, self.data_args, self.training_args, self.finetuning_args = prepare_args()
        if len(dataset_list) == 1:
            self.dataset = dataset_list[0].dataset
        else:
            self.dataset = concatenate_datasets([x.dataset for x in dataset_list])
        self.model, self.tokenizer = load_pretrained(self.model_args, self.finetuning_args, self.training_args.do_train)
        self.dataset = preprocess_data(template, self.dataset, self.tokenizer, self.data_args, self.training_args)
        self.data_collator = DynamicDataCollatorWithPadding(
            tokenizer=self.tokenizer,
            ignore_pad_token_for_loss=(self.data_args.ignore_pad_token_for_loss and not self.training_args.predict_with_generate)
        )

        # Override the decoding parameters of Seq2SeqTrainer
        self.training_args.generation_max_length = self.training_args.generation_max_length if \
            self.training_args.generation_max_length is not None else self.data_args.max_target_length
        self.training_args.generation_num_beams = self.data_args.eval_num_beams if \
            self.data_args.eval_num_beams is not None else self.training_args.generation_num_beams

        # Split the dataset
        if self.training_args.do_train:
            if self.data_args.dev_ratio > 1e-6:
                self.dataset, _ = self.dataset.train_test_split(test_size=self.data_args.dev_ratio)
                trainer_kwargs = {"train_dataset": self.dataset["train"], "eval_dataset": self.dataset["test"]}
            else:
                trainer_kwargs = {"train_dataset": self.dataset}
        else:  # do_eval or do_predict
            trainer_kwargs = {"eval_dataset": self.dataset}

        # Initialize our Trainer
        self.trainer = Seq2SeqPeftTrainer(
            finetuning_args=self.finetuning_args,
            model=self.model,
            args=self.training_args,
            tokenizer=self.tokenizer,
            data_collator=self.data_collator,
            callbacks=[LogCallback()],
            compute_metrics=ComputeMetrics(self.tokenizer) if self.training_args.predict_with_generate else None,
            **trainer_kwargs
        )

        # Keyword arguments for `model.generate`
        self.gen_kwargs = {
            "do_sample": True,
            "top_p": 0.7,
            "max_new_tokens": self.data_args.max_target_length + 1,
            "temperature": 0.95,
            "logits_processor": get_logits_processor()
        }

    def train(self):
        # Training
        if self.training_args.do_train:
            train_result = self.trainer.train()
            self.trainer.log_metrics("train", train_result.metrics)
            self.trainer.save_metrics("train", train_result.metrics)
            self.trainer.save_state()
            self.trainer.save_model()
            if self.trainer.is_world_process_zero() and self.model_args.plot_loss:
                plot_loss(self.training_args.output_dir, keys=["loss", "eval_loss"])

    def eval(self):
        # Evaluation
        if self.training_args.do_eval:
            metrics = self.trainer.evaluate(metric_key_prefix="eval", **self.gen_kwargs)
            if self.training_args.predict_with_generate:  # eval_loss will be wrong if predict_with_generate is enabled
                metrics.pop("eval_loss", None)
            self.trainer.log_metrics("eval", metrics)
            self.trainer.save_metrics("eval", metrics)

    def test(self):
        # Predict
        if self.training_args.do_predict:
            predict_results = self.trainer.predict(self.dataset, metric_key_prefix="predict", **self.gen_kwargs)
            if self.training_args.predict_with_generate:  # predict_loss will be wrong if predict_with_generate is enabled
                predict_results.metrics.pop("predict_loss", None)
            self.trainer.log_metrics("predict", predict_results.metrics)
            self.trainer.save_metrics("predict", predict_results.metrics)
            self.trainer.save_predictions(predict_results)



import json
import re
import collections
import string
import sys

def normalize_answer(s):
    """Lower text and remove punctuation, articles and extra whitespace."""

    def remove_articles(text):
        regex = re.compile(r"\b(a|an|the)\b", re.UNICODE)
        return re.sub(regex, " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def get_tokens(s):
    if not s:
        return []
    return normalize_answer(s).split()


def compute_exact(a_gold, a_pred):
    return int(normalize_answer(a_gold) == normalize_answer(a_pred))


def compute_f1(a_gold, a_pred):
    gold_toks = get_tokens(a_gold)
    pred_toks = get_tokens(a_pred)
    common = collections.Counter(gold_toks) & collections.Counter(pred_toks)
    num_same = sum(common.values())
    if len(gold_toks) == 0 or len(pred_toks) == 0:
        # If either is no-answer, then F1 is 1 if they agree, 0 otherwise
        return int(gold_toks == pred_toks)
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(pred_toks)
    recall = 1.0 * num_same / len(gold_toks)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1

def get_raw_scores(examples, reference):
    """
    Computes the exact and f1 scores from the examples and the model predictions
    """
    exact_scores = {}
    f1_scores = {}

    for example in examples:
        qas_id = example['question_id']
        gold_answers = [reference['reference'][qas_id]]

        prediction = example['pred']
        exact_scores[qas_id] = max(compute_exact(a, prediction) for a in gold_answers)
        f1_scores[qas_id] = max(compute_f1(a, prediction) for a in gold_answers)

    qid_list = reference['reference'].keys()
    total = len(qid_list)
    
    table_list = reference['table']
    passage_list = reference['passage']


########################################################################
    aa = sum([exact_scores[item] for item in exact_scores.keys()])
    bb = sum([f1_scores[item] for item in exact_scores.keys()])

    return collections.OrderedDict(
        [
            ("total exact", float(aa / len(exact_scores.keys()))   ),
            ("total f1", float(aa / len(f1_scores.keys()))),
            ("total", len(f1_scores.keys())),
        ]
    )
########################################################################
    
    


# assert len(sys.argv) == 3, "you need to input the file"

# with open(sys.argv[1], 'r') as f:
#     data = json.load(f)

# with open(sys.argv[2], 'r') as f:
#     ref = json.load(f)

# print(get_raw_scores(data, ref))



import math
import os
from io import BytesIO

from flask import Flask, render_template, jsonify, request, send_file, session
from flask_cors import CORS
import pandas as pd
import random
import yaml
import json
import glob
import shutil
import logging
import sys
import traceback
import pickle
from loaders import DATASET_CLASSES
from structs.data import Table, Cell
from utils import export
from try_table_qa_kit import TableQAKitDemo

def init_app():
    flask_app = Flask(
            "TextTableQA",
            template_folder=os.path.join(os.path.dirname(__file__), "templates"),
            static_folder=os.path.join(os.path.dirname(__file__), "static")
        )

    flask_app.database = {}
    flask_app.database['dataset'] = {}
    flask_app.config.update(SECRET_KEY=os.urandom(24))
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "config.yml")) as f:
        config = yaml.safe_load(f)
    flask_app.config.update(config)
    flask_app.pipeline = TableQAKitDemo(model_type="text_davinci_003", key="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")

    
    return flask_app

app = init_app()
CORS(app)


def init_log():
    log_format = "%(levelname)s:"
    logging.basicConfig(format=log_format, level=logging.INFO)
    file_handler = logging.FileHandler("error.log")
    file_handler.setLevel(logging.ERROR)
    logging.getLogger().addHandler(file_handler)
    logging.getLogger().addHandler(logging.StreamHandler())
    logger = logging.getLogger(__name__)
    return logger
logger = init_log()





'''
check data integrity
'''


def check_data_integrity(dataset_name, split, table_idx):
    check_dataset_in_database(dataset_name)
    check_data_in_dataset(dataset_name, split)
    check_table_in_dataset(dataset_name, split, table_idx)


def check_dataset_in_database(dataset_name):
    is_exist = dataset_name in app.database["dataset"]
    if not is_exist:
        logger.info(f"Creating Dataset Object {dataset_name}")
        create_dataset(dataset_name)


def create_dataset(dataset_name):
    dataset_obj = DATASET_CLASSES[dataset_name]()
    app.database['dataset'][dataset_name] = dataset_obj


def check_data_in_dataset(dataset_name, split):
    dataset_obj = app.database['dataset'][dataset_name]
    if not dataset_obj.has_split(split):
        logger.info(f"Loading {dataset_name} - {split}")
        dataset_obj.load(split=split)


def check_table_in_dataset(dataset_name, split, table_idx):
    dataset_obj = app.database['dataset'][dataset_name]
    is_exist = table_idx in dataset_obj.tables[split]
    if not is_exist:
        entry = dataset_obj.get_data(split, table_idx)
        table = dataset_obj.prepare_table(entry)
        dataset_obj.set_table(split, table_idx, table)

def statistics_default_table_information(dataset_name, split, table_idx, propertie_name_list):
    if dataset_name not in app.config['datasets']:
        raise Exception(f"datasets {dataset_name} not found")
    elif split not in app.config['split']:
        raise Exception(f"split {split} not found")
    check_data_integrity(dataset_name, split, table_idx)
    dataset_obj = app.database["dataset"][dataset_name]
    table_data = dataset_obj.get_table(split, table_idx)
    properties_html, table_html = export.table_to_html(table_data, propertie_name_list,None,"web")
    generated_results = fetch_generated_outputs(dataset_name, split, table_idx)
    data = {
        # "session": {},
        "table_cnt": dataset_obj.get_example_count(split),
        "generated_results": generated_results,
        "dataset_info": dataset_obj.get_info(),
        "table_question": table_data.default_question,
        "properties_html": properties_html,
        "table_html": table_html,
        "pictures": table_data.pic_info,  # 如果无，返回[]
        "text": {(index + 1): value for index, value in enumerate(table_data.txt_info)},  # 如果无，返回{}
        "success": True
    }
    # print("table_cnt\n", dataset_obj.get_example_count(split))
    # print("generated_results\n", generated_results)
    # print("dataset_info\n", dataset_obj.get_info())
    # print("table_question\n", table_data.default_question)
    # with open("properties.html", "w", encoding="utf-8") as file:
    #     file.write(properties_html)
    # with open("table.html", "w", encoding="utf-8") as file:
    #     file.write(table_html)
    # print("pictures\n", table_data.pic_info)
    # print("text\n", {(index + 1): value for index, value in enumerate(table_data.txt_info)})
    return data
# done
@app.route("/default/table", methods=["GET", "POST"])
def fetch_default_table_data():
    try:
        content = request.json
        dataset_name = content.get("dataset_name")
        split = content.get("split")
        table_idx = content.get("table_idx")

        propertie_name_list = []
        data = statistics_default_table_information(dataset_name, split, table_idx, propertie_name_list)
        return jsonify(data),200

    except Exception as e:
        logger.error(f"Fetch Table Error: {e}")
        data = {"success": False}
        return jsonify(data),400 


def fetch_generated_outputs(dataset_name, split, table_idx):
    
    outputs = {}

    model_list = os.listdir(f"outputs/{dataset_name}")
    valid_model_list = []
    for model_name in model_list:
        split_list = os.listdir(f"outputs/{dataset_name}/{model_name}")
        for split_file_name in split_list:
            split_name, _ = os.path.splitext(split_file_name)
            if split == split_name:
                valid_model_list.append(model_name)
                break
    for valid_model_name in valid_model_list:
        outputs_file_path = f'outputs/{dataset_name}/{valid_model_name}/{split}.txt'
        with open(outputs_file_path, 'r', encoding='utf-8') as file:
            lines = file.readlines()
            if table_idx < len(lines):
                outputs[valid_model_name] = lines[table_idx].strip()        
    return outputs

# done
@app.route("/custom/table", methods=["GET", "POST"])
def fetch_custom_table_data():
    try:
        json_file = request.json
        table_name = json_file.get("table_name")
        # table_name = "我的自定义表格1"
        properties_name_list = []
        custom_tables = session.get("custom_tables", {})
        if len(custom_tables) != 0 and table_name in custom_tables:
            table_data = pickle.loads(custom_tables[table_name])
            properties_html,table_html = export.table_to_html(table_data, properties_name_list,None,"web")
            data = {
                "table_html": table_html,
                "properties_html": properties_html,
                "success": True
            }
            # with open("properties.html", "w", encoding="utf-8") as file:
            #     file.write(properties_html)
            # with open("table.html", "w", encoding="utf-8") as file:
            #     file.write(table_html)
            return jsonify(data),200
        else:
            raise Exception("fetch non-existent tables in session")
    except Exception as e:
        logger.error(f"Fetch Table Error: {e}")
        data = {"success": False}
        return jsonify(data),400

# done
@app.route("/custom/remove", methods=["GET", "POST"])
def remove_custom_table():
    try:
        json_file = request.json
        table_name = json_file.get("table_name")
        # table_name = "我的自定义表格1"
        custom_tables = session.get("custom_tables", {})
        if len(custom_tables) != 0 and table_name in custom_tables:
            custom_tables.pop(table_name)
        else:
            raise Exception("delete non-existent tables in session")
        session["custom_tables"] = custom_tables
        session.modified = True
        result = {"success" : True}
        return jsonify(result),200
    except Exception as e:
        logger.error(f"Remove Table Error: {e}")
        result = {"success" : False}
        return jsonify(result),400

# done
@app.route("/session", methods=["GET", "POST"])
def get_session():
    try:
        json_file = request.json
        target = json_file.get("target")
        # target = "custom_tables_name"
        if target == "custom_tables_name":
            data = {
                "data" : list(session.get("custom_tables", {}).keys()), #["a","b"]
                "success" : True
            }
            return jsonify(data),200
        else:
            raise Exception("Illegal Target")
    except Exception as e:
        logger.error(f"Get Session Error: {e}")
        data = {"success" : False}
        return jsonify(data),400
    


@app.route("/default/pipeline", methods=["GET", "POST"])
def fetch_default_pipeline_result():
    try:
        content = request.json
        dataset_name = content.get("dataset_name")
        split = content.get("split")
        table_idx = content.get("table_idx")
        question = content.get("question")


        if dataset_name not in app.config['datasets']:
            raise Exception(f"datasets {dataset_name} not found")
        elif split not in app.config['split']:
            raise Exception(f"split {split} not found")

        answer = app.pipeline.try_table_qa_kit_by_id(data=dataset_name, dataset_type=split, index=table_idx, question=question)
        data = {
            "answer": answer,
            "success": True
        }
        return jsonify(data),200
        
    except Exception as e:
        logger.error(f"Pipeline Error: {e}")
        traceback.print_exc()
        data = {"success": False}
        return jsonify(data),400

@app.route("/custom/pipeline", methods=["GET", "POST"])
def fetch_custom_pipeline_result():
    try:
        content = request.json
        table_name = content.get("table_name")
        question = content.get("question")

        custom_tables = session.get("custom_tables", {})
        if len(custom_tables) != 0 and table_name in custom_tables:
            table_data = pickle.loads(custom_tables[table_name])
            cells = table_data.cells
            input = [[cell.value for cell in row] for row in cells]
            answer = app.pipeline.try_table_qa_kit(table=input, texts=None, question=question)

            data = {
                "answer": answer,
                "success": True
            }

            return jsonify(data),200
        else:
            raise Exception("fetch non-existent tables in session")
    except Exception as e:
        logger.error(f"Pipeline Error: {e}")
        data = {"success": False}
        return jsonify(data),400


# done
@app.route("/custom/upload", methods=["GET", "POST"])
def upload_custom_table():
    # 上传的表格名不能重复,前端校验+后端校验 限制文件大小
    #############################################################
    try:
        if 'excel_file' in request.files:
            file = request.files['excel_file']
        else:
            raise Exception("no excel file")
        # json_file = request.json
        # table_name = json_file.get('table_name')
    
        table_name, file_extension = os.path.splitext(file.filename)

        # file = 'test.xlsx'
        properties = {}
        # properties = {"title": "List of Governors of South Carolina", "overlap_subset": "True"}

        # table_name = "我的自定义表格1"

        custom_tables = session.get("custom_tables", {})
        if table_name in custom_tables:
            raise Exception("add duplicate names to tables in session")
        else:     
            if file_extension == '.xls':
                df = pd.read_excel(file, dtype=str, engine="xlrd")
            elif file_extension == '.xlsx':
                df = pd.read_excel(file, dtype=str, engine="openpyxl")
            else:
                raise Exception("Unsupported file format. Only .xls and .xlsx are supported.")

            headers = df.columns.tolist()
            data = df.values.tolist()
            table_data = prepare_custom_table(headers, data, properties, table_name)
            custom_tables[table_name] = pickle.dumps(table_data)
        session["custom_tables"] = custom_tables
        session.modified = True
        result = {"success" : True}
        return jsonify(result),200
    except Exception as e:
        logger.error(f"Upload Table Error: {e}")
        result = {"success" : False}
        return jsonify(result),400

    


def prepare_custom_table(headers, data, properties, table_name):
    t = Table()
    t.type = "custom"
    t.custom_table_name = table_name

    for key in properties:
        t.props[key] = properties[key]

    for header_cell in headers:
        c = Cell()
        # if math.isnan(header_cell):
        #     header_cell = ""
        if isinstance(header_cell, float) and math.isnan(header_cell):
            header_cell = ""
        c.value = header_cell
        c.is_col_header = True
        t.add_cell(c)
    t.save_row()

    for row in data:
        for cell in row:
            c = Cell()
            # if math.isnan(cell):
            #     cell = ""
            if isinstance(cell, float) and math.isnan(cell):
                cell = ""
            c.value = cell
            t.add_cell(c)
        t.save_row()

    return t
def download_table(format, table_data, include_props, file_name):
    if format == "txt":
        content = export.table_to_linear(table_data, include_props)
        file_stream = BytesIO(content.encode('utf-8'))
        file_stream.seek(0)
        return send_file(
            file_stream,
            mimetype="text/plain",
            download_name=f"{file_name}.{format}",
            as_attachment=True
        ),200
    elif format == "json":
        content = export.table_to_json(table_data, include_props)
        json_data = json.dumps(content)
        file_stream = BytesIO(json_data.encode('utf-8'))
        file_stream.seek(0)
        return send_file(
            file_stream,
            mimetype='application/json',
            download_name=f"{file_name}.{format}",
            as_attachment=True
        ),200

    elif format == "xlsx":
        file_stream = export.table_to_excel(table_data, include_props)
        file_stream.seek(0)
        return send_file(
            file_stream,
            mimetype='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
            download_name=f"{file_name}.{format}",
            as_attachment=True,
        ),200

    elif format == "csv":
        content = export.table_to_csv(table_data)
        file_stream = BytesIO(content.encode('utf-8'))
        file_stream.seek(0)
        return send_file(
            file_stream,
            mimetype="text/csv",
            download_name=f"{file_name}.{format}",
            as_attachment=True
        ),200
    elif format == "html":
        content = export.table_to_html(table_data, None, include_props, "export", merge=True)
        file_stream = BytesIO(content.encode('utf-8'))
        file_stream.seek(0)
        return send_file(
            file_stream,
            mimetype="text/html",
            download_name=f"{file_name}.{format}",
            as_attachment=True
        ),200
    else:
        raise Exception("illegal format")
# done
@app.route("/custom/download", methods=["GET", "POST"])
def download_custom_table():
    try:
        format = request.args.get('format')
        table_name = request.args.get('table_name')
        # format = "json"
        include_props = False
        # table_name = "我的自定义表格1"

        custom_tables = session.get("custom_tables", {})
        if len(custom_tables) != 0 and table_name in custom_tables:
            table_data = pickle.loads(custom_tables[table_name])
            return download_table(format, table_data, include_props, f"custom_{table_name}")
        else:
            raise Exception("download non-existent tables in session")
    except Exception as e:
        logger.error(f"Download Table Error: {e}")
        return jsonify({"success" : False}),400
# done
@app.route("/default/download", methods=["GET", "POST"])
def download_default_table():
    try:
        format = request.args.get('format')
        include_props = request.args.get('include_props')
        dataset_name = request.args.get('dataset_name')
        split = request.args.get('split')
        table_idx = int(request.args.get('table_idx'))


        if dataset_name not in app.config['datasets']:
            raise Exception(f"datasets {dataset_name} not found")
        elif split not in app.config['split']:
            raise Exception(f"split {split} not found")
        elif dataset_name == "MultiHiertt" and format != "html":
            raise Exception(f"dataset {dataset_name} can only be downloaded in HTML format")
        check_data_integrity(dataset_name, split, table_idx)
        dataset_obj = app.database["dataset"][dataset_name]
        table_data = dataset_obj.get_table(split, table_idx)
        return download_table(format, table_data, include_props, f"{dataset_name}_{split}_{table_idx}")
    except Exception as e:
        logger.error(f"Download Table Error: {e}")
        logger.error(e.__traceback__)
        return jsonify({"success" : False}),400
# done
@app.route("/custom", methods=["GET", "POST"])
def custom_mode():
    return render_template("custom_mode.html")
# done
@app.route("/custom/example", methods=["GET", "POST"])
def download_file_example():
    try:
        file_response = send_file(
            app.config['example_filename'],
            as_attachment=True,
            mimetype='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
            download_name="example.xlsx"
        ),200
        return file_response
    except Exception as e:
        logger.error(f"Download Example Error: {e}")
        return jsonify({"success" : False}),400

# done
@app.route("/", methods=["GET", "POST"])
def index():

    # try:
    #     dataset_name = app.config['default_dataset']
    #     split = "train"
    #     table_idx = 0
    #     propertie_name_list = []
    #     logger.info(f"Page loaded")
    #     data = statistics_default_table_information(dataset_name, split, table_idx, propertie_name_list)
    # except Exception as e:
    #     logger.error(f"Fetch initial Table Error: {e}")
    #     data = {"success": False}


    return render_template(
        "index.html"
        # table_data = data  #--这样传没有jsonfiy应该是可以的--
    )

with app.app_context():
    print("Initing datasets")
    for dataset_name in app.config['datasets']:
        for split in app.config['split']:
            check_data_integrity(dataset_name, split, 0)
    print("done")
#     import psutil
#     # 获取内存使用情况
#     memory_info = psutil.virtual_memory()

#     # 打印内存总量、使用量、可用量等信息
#     print(f"1Total Memory: {memory_info.total} bytes")
#     print(f"1Used Memory: {memory_info.used} bytes")
#     print(f"1Available Memory: {memory_info.available} bytes")
#     print(f"1Percentage Used: {memory_info.percent}%")
#     print("load start")
#     for dataset_name in app.config['datasets']:
#         for split in app.config['split']:
#             check_data_integrity(dataset_name, split, 0)
#     print("load over")
#     # 获取内存使用情况
#     memory_info = psutil.virtual_memory()

#     # 打印内存总量、使用量、可用量等信息
#     print(f"2Total Memory: {memory_info.total} bytes")
#     print(f"2Used Memory: {memory_info.used} bytes")
#     print(f"2Available Memory: {memory_info.available} bytes")
#     print(f"2Percentage Used: {memory_info.percent}%")



    # print(fetch_generated_outputs("spreadsheetqa","dev",10))



#     fetch_table_data()
#     dataset_obj = app.database['dataset']["wikisql"]
#     print(dataset_obj.tables)
#     session1 = {}
#     upload_custom_table()
#     fetch_custom_table_data()
#     session["custom_tables"] = {}
#     fetch_default_table_data()
#     pass
#     download_default_table()
#     upload_custom_table()
# if __name__ == '__main__':
#     app.run(host = "210.75.240.136", port = 18888)



import os

import datasets
import logging
import json
from structs.data import Cell, Table, HFTabularDataset


logger = logging.getLogger(__name__)

class MultiModalQA(HFTabularDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _load_split(self, split):

        question_file_map = {
            "train" : "multimodalqa_final_dataset_pipeline_camera_ready_MMQA_train.jsonl",
            "dev" : "multimodalqa_final_dataset_pipeline_camera_ready_MMQA_dev.jsonl",
            "test" : "multimodalqa_final_dataset_pipeline_camera_ready_MMQA_test.jsonl"
        }
        logger.info(f"Loading MultimodalQA - {split}")


        '''
        loading dataset
        '''
        # loading txt
        txt_file_path = "datasets/MultimodalQA/multimodalqa_final_dataset_pipeline_camera_ready_MMQA_texts.jsonl"
        txt_info = {}

        with open(txt_file_path, 'r', encoding='utf-8') as file:
            for line in file:
                json_data = json.loads(line)
                id_value = json_data['id']
                text_value = json_data['text']
                txt_info[id_value] = text_value

        # print(txt_info)

        #loading pic
        pic_file_path = "datasets/MultimodalQA/multimodalqa_final_dataset_pipeline_camera_ready_MMQA_images.jsonl"
        pic_info = {}

        with open(pic_file_path, 'r', encoding='utf-8') as file:
            for line in file:
                json_data = json.loads(line)
                id_value = json_data['id']
                path_value = json_data['path']
                pic_info[id_value] = path_value

        # print(pic_info)

        #loading table
        table_file_path = "datasets/MultimodalQA/multimodalqa_final_dataset_pipeline_camera_ready_MMQA_tables.jsonl"
        table_info = {}
        table_meta_info = {}

        with open(table_file_path, 'r', encoding='utf-8') as file:
            for line in file:
                json_data = json.loads(line)
                id_value = json_data['id']
                table_value = json_data['table']
                table_rows_value = table_value["table_rows"]
                table_rows_text_value = []
                for rows in table_rows_value:
                    temp_list = []
                    for col in rows:
                        temp_list.append(col["text"])
                    table_rows_text_value.append(temp_list)
                header_value = table_value["header"]
                header_text_value = []
                for row in header_value:
                    header_text_value.append(row["column_name"])
                table_info[id_value] = {}
                table_meta_info[id_value] = {}
                table_info[id_value]["table_rows_text"] = table_rows_text_value
                table_info[id_value]["header_text"] = header_text_value
                if 'title' in json_data:
                    table_meta_info[id_value]["title"] = json_data['title']
                if 'url' in json_data:
                    table_meta_info[id_value]["url"] = json_data['url']
                if 'table_name' in json_data['table']:
                    table_meta_info[id_value]["table_name"] = json_data['table']['table_name']
        # print(table_info)

        # loading question
        question_file_name = question_file_map[split]
        question_file_path = os.path.join("datasets/MultimodalQA", question_file_name)
        question_info = []
        # print(question_file_path)

        with open(question_file_path, 'r', encoding='utf-8') as file:
            for line in file:
                properties_info = {}
                json_data = json.loads(line)
                # qid_value = json_data['qid']
                question_value = json_data['question']
                image_ids_value = json_data['metadata']['image_doc_ids']
                txt_ids_value = json_data['metadata']['text_doc_ids']
                table_id_value = json_data['metadata']["table_id"]
                if 'type' in json_data['metadata']:
                    properties_info['question--type'] = str(json_data['metadata']['type'])
                if 'modalities' in json_data['metadata']:
                    properties_info['question--modalities'] = str(json_data['metadata']['modalities'])
                if 'wiki_entities_in_answers' in json_data['metadata']:
                    properties_info['question--wiki_entities_in_answers'] = str(json_data['metadata']['wiki_entities_in_answers'])
                if 'wiki_entities_in_question' in json_data['metadata']:
                    properties_info['question--wiki_entities_in_question'] = str(json_data['metadata']['wiki_entities_in_question'])
                if 'pseudo_language_question' in json_data['metadata']:
                    properties_info['question--pseudo_language_question'] = str(json_data['metadata']['pseudo_language_question'])
                for key in table_meta_info[table_id_value]:
                    properties_info[f'table--{key}'] = str(table_meta_info[table_id_value][key])
                question_info.append({
                    "question" : question_value,
                    "image" : [pic_info[i] for i in image_ids_value],
                    "txt" : [txt_info[i] for i in txt_ids_value],
                    "table": table_info[table_id_value],
                    "properties" : properties_info
                })

        self.dataset_info = {
            "citation" : "111",
            "description" : "111",
            "version" : "111",
            "license" : "111"
        }
        self.data[split] = question_info

    def load_split_test(self, split):
        self._load_split(split)

    def prepare_table(self, entry):
        t = Table()
        t.type = "default"
        t.default_question = entry["question"]
        t.pic_info = entry["image"]
        t.txt_info = entry["txt"]

        t.props = entry['properties']

        for header_cell in entry["table"]["header_text"]:
            c = Cell()
            c.value = header_cell
            c.is_col_header = True
            t.add_cell(c)
        t.save_row()

        for row in entry["table"]["table_rows_text"]:
            for cell in row:
                c = Cell()
                c.value = cell
                t.add_cell(c)
            t.save_row()
        return t
if __name__ == '__main__':
    MultiModalQA().load_split_test("train")


import json
import logging
import os
import sys
from dataclasses import dataclass, field, asdict
from typing import Optional, Literal, Dict, Any, Tuple

import datasets
import torch
import transformers
from transformers import HfArgumentParser, Seq2SeqTrainingArguments, LogitsProcessorList, LogitsProcessor

IGNORE_INDEX = -100
VALUE_HEAD_FILE_NAME = "value_head.bin"
FINETUNING_ARGS_NAME = "finetuning_args.json"


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune.
    """
    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models."}
    )
    cache_dir: Optional[str] = field(
        default=None,
        metadata={"help": "Where to store the pretrained models downloaded from huggingface.co."}
    )
    use_fast_tokenizer: Optional[bool] = field(
        default=False,
        metadata={"help": "Whether to use one of the fast tokenizer (backed by the tokenizers library) or not."}
    )
    use_auth_token: Optional[bool] = field(
        default=False,
        metadata={"help": "Will use the token generated when running `huggingface-cli login`."}
    )
    model_revision: Optional[str] = field(
        default="main",
        metadata={"help": "The specific model version to use (can be a branch name, tag name or commit id)."}
    )
    quantization_bit: Optional[int] = field(
        default=None,
        metadata={"help": "The number of bits to quantize the model."}
    )
    quantization_type: Optional[Literal["fp4", "nf4"]] = field(
        default="nf4",
        metadata={"help": "Quantization data type to use in int4 training."}
    )
    double_quantization: Optional[bool] = field(
        default=True,
        metadata={"help": "Whether to use double quantization in int4 training or not."}
    )
    compute_dtype: Optional[torch.dtype] = field(
        default=None,
        metadata={"help": "Used in quantization configs. Do not specify this argument manually."}
    )
    checkpoint_dir: Optional[str] = field(
        default=None,
        metadata={"help": "Path to the directory(s) containing the delta model checkpoints as well as the configurations."}
    )
    resume_lora_training: Optional[bool] = field(
        default=True,
        metadata={"help": "Whether to resume training from the last LoRA weights or create new weights after merging them."}
    )
    plot_loss: Optional[bool] = field(
        default=False,
        metadata={"help": "Whether to plot the training loss after fine-tuning or not."}
    )

    def __post_init__(self):
        if self.checkpoint_dir is not None:  # support merging multiple lora weights
            self.checkpoint_dir = [cd.strip() for cd in self.checkpoint_dir.split(",")]

        if self.quantization_bit is not None:
            assert self.quantization_bit in [4, 8], "We only accept 4-bit or 8-bit quantization."


@dataclass
class DataTrainingArguments:
    """
    Arguments pertaining to what data we are going to input our model for training and evaluation.
    """
    overwrite_cache: Optional[bool] = field(
        default=False,
        metadata={"help": "Overwrite the cached training and evaluation sets."}
    )
    preprocessing_num_workers: Optional[int] = field(
        default=None,
        metadata={"help": "The number of processes to use for the preprocessing."}
    )
    max_source_length: Optional[int] = field(
        default=512,
        metadata={"help": "The maximum total input sequence length after tokenization."}
    )
    max_target_length: Optional[int] = field(
        default=512,
        metadata={"help": "The maximum total output sequence length after tokenization."}
    )
    max_samples: Optional[int] = field(
        default=None,
        metadata={"help": "For debugging purposes, truncate the number of examples for each dataset."}
    )
    eval_num_beams: Optional[int] = field(
        default=None,
        metadata={"help": "Number of beams to use for evaluation. This argument will be passed to `model.generate`"}
    )
    ignore_pad_token_for_loss: Optional[bool] = field(
        default=True,
        metadata={"help": "Whether to ignore the tokens corresponding to padded labels in the loss computation or not."}
    )
    dev_ratio: Optional[float] = field(
        default=0,
        metadata={"help": "Proportion of the dataset to include in the development set, should be between 0.0 and 1.0."}
    )


@dataclass
class FinetuningArguments:
    """
    Arguments pertaining to which techniques we are going to fine-tuning with.
    """
    finetuning_type: Optional[Literal["none", "freeze", "lora", "full"]] = field(
        default="lora",
        metadata={"help": "Which fine-tuning method to use."}
    )
    num_hidden_layers: Optional[int] = field(
        default=32,
        metadata={"help": "Number of decoder blocks in the model. \
                  LLaMA choices: [\"32\", \"40\", \"60\", \"80\"], \
                  BLOOM choices: [\"24\", \"30\", \"70\"], \
                  Baichuan choices: [\"32\"]"}
    )
    num_layer_trainable: Optional[int] = field(
        default=3,
        metadata={"help": "Number of trainable layers for Freeze fine-tuning."}
    )
    name_module_trainable: Optional[Literal["mlp", "self_attn", "self_attention"]] = field(
        default="mlp",
        metadata={"help": "Name of trainable modules for Freeze fine-tuning. \
                  LLaMA choices: [\"mlp\", \"self_attn\"], \
                  BLOOM choices: [\"mlp\", \"self_attention\"], \
                  Baichuan choices: [\"mlp\", \"self_attn\"]"}
    )
    lora_rank: Optional[int] = field(
        default=8,
        metadata={"help": "The intrinsic dimension for LoRA fine-tuning."}
    )
    lora_alpha: Optional[int] = field(
        default=32.0,
        metadata={"help": "The scale factor for LoRA fine-tuning (similar with the learning rate)."}
    )
    lora_dropout: Optional[float] = field(
        default=0.1,
        metadata={"help": "Dropout rate for the LoRA fine-tuning."}
    )
    lora_target: Optional[str] = field(
        default="q_proj,v_proj",
        metadata={"help": "Name(s) of target modules to apply LoRA. Use comma to separate multiple modules. \
                  LLaMA choices: [\"q_proj\", \"k_proj\", \"v_proj\", \"o_proj\", \"gate_proj\", \"up_proj\", \"down_proj\"], \
                  BLOOM choices: [\"query_key_value\", \"self_attention.dense\", \"mlp.dense\"], \
                  Baichuan choices: [\"W_pack\", \"o_proj\", \"gate_proj\", \"up_proj\", \"down_proj\"]"}
    )

    def __post_init__(self):
        if isinstance(self.lora_target, str): # support custom target modules/layers of LoRA
            self.lora_target = [target.strip() for target in self.lora_target.split(",")]

        if self.num_layer_trainable > 0: # fine-tuning the last n layers if num_layer_trainable > 0
            trainable_layer_ids = [self.num_hidden_layers - k for k in range(self.num_layer_trainable)]
        else: # fine-tuning the first n layers if num_layer_trainable < 0
            trainable_layer_ids = [k for k in range(-self.num_layer_trainable)]

        self.trainable_layers = ["{:d}.{}".format(idx, self.name_module_trainable) for idx in trainable_layer_ids]

        assert self.finetuning_type in ["none", "freeze", "lora", "full"], "Invalid fine-tuning method."

    def save_to_json(self, json_path: str):
        """Saves the content of this instance in JSON format inside `json_path`."""
        json_string = json.dumps(asdict(self), indent=2, sort_keys=True) + "\n"
        with open(json_path, "w", encoding="utf-8") as f:
            f.write(json_string)

    @classmethod
    def load_from_json(cls, json_path: str):
        """Creates an instance from the content of `json_path`."""
        with open(json_path, "r", encoding="utf-8") as f:
            text = f.read()
        return cls(**json.loads(text))


@dataclass
class GeneratingArguments:
    """
    Arguments pertaining to specify the decoding parameters.
    """
    do_sample: Optional[bool] = field(
        default=True,
        metadata={"help": "Whether or not to use sampling, use greedy decoding otherwise."}
    )
    temperature: Optional[float] = field(
        default=0.95,
        metadata={"help": "The value used to modulate the next token probabilities."}
    )
    top_p: Optional[float] = field(
        default=0.7,
        metadata={"help": "The smallest set of most probable tokens with probabilities that add up to top_p or higher are kept."}
    )
    top_k: Optional[int] = field(
        default=50,
        metadata={"help": "The number of highest probability vocabulary tokens to keep for top-k filtering."}
    )
    num_beams: Optional[int] = field(
        default=1,
        metadata={"help": "Number of beams for beam search. 1 means no beam search."}
    )
    max_length: Optional[int] = field(
        default=None,
        metadata={"help": "The maximum length the generated tokens can have. It can be overridden by max_new_tokens."}
    )
    max_new_tokens: Optional[int] = field(
        default=512,
        metadata={"help": "The maximum numbers of tokens to generate, ignoring the number of tokens in the prompt."}
    )
    repetition_penalty: Optional[float] = field(
        default=1.0,
        metadata={"help": "The parameter for repetition penalty. 1.0 means no penalty."}
    )
    length_penalty: Optional[float] = field(
        default=1.0,
        metadata={"help": "Exponential penalty to the length that is used with beam-based generation."}
    )

    def to_dict(self) -> Dict[str, Any]:
        args = asdict(self)
        if args.get("max_new_tokens", None):
            args.pop("max_length", None)
        return args


def get_logger(name: str) -> logging.Logger:
    return logging.getLogger(name)


logger = get_logger(__name__)


def prepare_args() -> Tuple[ModelArguments, DataTrainingArguments, Seq2SeqTrainingArguments, FinetuningArguments]:

    parser = HfArgumentParser((ModelArguments, DataTrainingArguments, Seq2SeqTrainingArguments, FinetuningArguments))

    if len(sys.argv) == 2 and sys.argv[1].endswith(".json"): # Provide arguments with a json file.
        model_args, data_args, training_args, finetuning_args = parser.parse_json_file(json_file=os.path.abspath(sys.argv[1]))
    else:
        model_args, data_args, training_args, finetuning_args = parser.parse_args_into_dataclasses()

    # Setup logging
    if training_args.should_log:
        # The default of training_args.log_level is passive, so we set log level at info here to have that default.
        transformers.utils.logging.set_verbosity_info()

    log_level = training_args.get_process_log_level()
    datasets.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.enable_default_handler()
    transformers.utils.logging.enable_explicit_format()

    assert not (training_args.do_train and training_args.predict_with_generate), \
        "`predict_with_generate` cannot be set as True while training."

    assert (not training_args.do_predict) or training_args.predict_with_generate, \
        "Please enable `predict_with_generate` to save model predictions."

    assert model_args.quantization_bit is None or finetuning_args.finetuning_type == "lora", \
        "Quantization is only compatible with the LoRA method."

    if model_args.checkpoint_dir is not None:
        if finetuning_args.finetuning_type != "lora":
            assert len(model_args.checkpoint_dir) == 1, "Only LoRA tuning accepts multiple checkpoints."
        else:
            assert model_args.quantization_bit is None or len(model_args.checkpoint_dir) == 1, \
                "Quantized model only accepts a single checkpoint."

    if model_args.quantization_bit is not None and (not training_args.do_train):
        logger.warning("Evaluating model in 4/8-bit mode may cause lower scores.")

    if training_args.do_train and (not training_args.fp16):
        logger.warning("We recommend enable fp16 mixed precision training.")

    if training_args.local_rank != -1 and training_args.ddp_find_unused_parameters is None:
        logger.warning("`ddp_find_unused_parameters` needs to be set as False in DDP training.")
        training_args.ddp_find_unused_parameters = False

    training_args.optim = "adamw_torch" if training_args.optim == "adamw_hf" else training_args.optim # suppress warning

    if model_args.quantization_bit is not None:
        if training_args.fp16:
            model_args.compute_dtype = torch.float16
        elif training_args.bf16:
            model_args.compute_dtype = torch.bfloat16
        else:
            model_args.compute_dtype = torch.float32

    # Log on each process the small summary:
    logger.info(
        f"Process rank: {training_args.local_rank}, device: {training_args.device}, n_gpu: {training_args.n_gpu}\n"
        + f"  distributed training: {bool(training_args.local_rank != -1)}, 16-bits training: {training_args.fp16}"
    )
    logger.info(f"Training/evaluation parameters {training_args}")

    # Set seed before initializing model.
    transformers.set_seed(training_args.seed)

    return model_args, data_args, training_args, finetuning_args


def prepare_infer_args() -> Tuple[ModelArguments, DataTrainingArguments, FinetuningArguments, GeneratingArguments]:

    parser = HfArgumentParser((ModelArguments, DataTrainingArguments, FinetuningArguments, GeneratingArguments))

    if len(sys.argv) == 2 and sys.argv[1].endswith(".json"): # Provide arguments with a json file.
        model_args, data_args, finetuning_args, generating_args = parser.parse_json_file(json_file=os.path.abspath(sys.argv[1]))
    else:
        model_args, data_args, finetuning_args, generating_args = parser.parse_args_into_dataclasses()

    assert model_args.quantization_bit is None or finetuning_args.finetuning_type == "lora", \
        "Quantization is only compatible with the LoRA method."

    if model_args.checkpoint_dir is not None:
        if finetuning_args.finetuning_type != "lora":
            assert len(model_args.checkpoint_dir) == 1, "Only LoRA tuning accepts multiple checkpoints."
        else:
            assert model_args.quantization_bit is None or len(model_args.checkpoint_dir) == 1, \
                "Quantized model only accepts a single checkpoint."

    return model_args, data_args, finetuning_args, generating_args


class InvalidScoreLogitsProcessor(LogitsProcessor):

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        if torch.isnan(scores).any() or torch.isinf(scores).any():
            scores.zero_()
            scores[..., 0] = 1.0
        return scores


def get_logits_processor() -> LogitsProcessorList:
    logits_processor = LogitsProcessorList()
    logits_processor.append(InvalidScoreLogitsProcessor())
    return logits_processor
